"""Useful functions for formatting outputs.
"""

def format_floats(data):
    """Returns formatted list of floats from unformatted list of floats

    Parameters
    ----------
    data : list 
        The list of floats to be formatted

    Returns
    -------
    list
        Formatted floats (now string).
    """
    return ['{0:.3f}'.format(elem) for elem in data]


def format_float(value):
    """ Formats a float.

    Parameters
    ----------
    value : float
        Float to be formatted

    Returns
    -------
    str 
        Formatted float.

    """
    return format_floats([value])[0]
    



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer_ui_files/singleChannelDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_singleChannelDialog(object):
    def setupUi(self, singleChannelDialog):
        singleChannelDialog.setObjectName("singleChannelDialog")
        singleChannelDialog.resize(419, 523)
        self.gridLayout = QtWidgets.QGridLayout(singleChannelDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(singleChannelDialog)
        self.scrollArea.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 399, 472))
        self.scrollAreaWidgetContents.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 4, 0, 1, 1)
        self.groupBoxLegend = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxLegend.setObjectName("groupBoxLegend")
        self.formLayoutLegend = QtWidgets.QFormLayout(self.groupBoxLegend)
        self.formLayoutLegend.setObjectName("formLayoutLegend")
        self.gridLayout_2.addWidget(self.groupBoxLegend, 3, 0, 1, 1)
        self.groupBoxValueRange = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxValueRange.setObjectName("groupBoxValueRange")
        self.formLayout_3 = QtWidgets.QFormLayout(self.groupBoxValueRange)
        self.formLayout_3.setObjectName("formLayout_3")
        self.labelMax = QtWidgets.QLabel(self.groupBoxValueRange)
        self.labelMax.setObjectName("labelMax")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelMax)
        self.doubleSpinBoxMax = QtWidgets.QDoubleSpinBox(self.groupBoxValueRange)
        self.doubleSpinBoxMax.setSuffix("")
        self.doubleSpinBoxMax.setDecimals(4)
        self.doubleSpinBoxMax.setMinimum(-100000000000.0)
        self.doubleSpinBoxMax.setMaximum(100000000000.0)
        self.doubleSpinBoxMax.setSingleStep(0.01)
        self.doubleSpinBoxMax.setProperty("value", 0.0)
        self.doubleSpinBoxMax.setObjectName("doubleSpinBoxMax")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxMax)
        self.labelMin = QtWidgets.QLabel(self.groupBoxValueRange)
        self.labelMin.setObjectName("labelMin")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.labelMin)
        self.doubleSpinBoxMin = QtWidgets.QDoubleSpinBox(self.groupBoxValueRange)
        self.doubleSpinBoxMin.setSuffix("")
        self.doubleSpinBoxMin.setDecimals(4)
        self.doubleSpinBoxMin.setMinimum(-100000000000.0)
        self.doubleSpinBoxMin.setMaximum(100000000000.0)
        self.doubleSpinBoxMin.setSingleStep(0.01)
        self.doubleSpinBoxMin.setProperty("value", 0.0)
        self.doubleSpinBoxMin.setObjectName("doubleSpinBoxMin")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxMin)
        self.gridLayout_2.addWidget(self.groupBoxValueRange, 2, 0, 1, 1)
        self.groupBoxGeneral = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxGeneral.setObjectName("groupBoxGeneral")
        self.formLayout_4 = QtWidgets.QFormLayout(self.groupBoxGeneral)
        self.formLayout_4.setObjectName("formLayout_4")
        self.labelChannel = QtWidgets.QLabel(self.groupBoxGeneral)
        self.labelChannel.setObjectName("labelChannel")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelChannel)
        self.comboBoxChannel = QtWidgets.QComboBox(self.groupBoxGeneral)
        self.comboBoxChannel.setObjectName("comboBoxChannel")
        self.formLayout_4.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.comboBoxChannel)
        self.labelTitle = QtWidgets.QLabel(self.groupBoxGeneral)
        self.labelTitle.setObjectName("labelTitle")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.labelTitle)
        self.lineEditTitle = QtWidgets.QLineEdit(self.groupBoxGeneral)
        self.lineEditTitle.setObjectName("lineEditTitle")
        self.formLayout_4.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEditTitle)
        self.gridLayout_2.addWidget(self.groupBoxGeneral, 0, 0, 1, 1)
        self.groupBoxSmoothing = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxSmoothing.setObjectName("groupBoxSmoothing")
        self.formLayout = QtWidgets.QFormLayout(self.groupBoxSmoothing)
        self.formLayout.setObjectName("formLayout")
        self.labelWindowLength = QtWidgets.QLabel(self.groupBoxSmoothing)
        self.labelWindowLength.setObjectName("labelWindowLength")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelWindowLength)
        self.spinBoxWindowLength = QtWidgets.QSpinBox(self.groupBoxSmoothing)
        self.spinBoxWindowLength.setMaximum(10000000)
        self.spinBoxWindowLength.setProperty("value", 11)
        self.spinBoxWindowLength.setObjectName("spinBoxWindowLength")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.spinBoxWindowLength)
        self.labelWindow = QtWidgets.QLabel(self.groupBoxSmoothing)
        self.labelWindow.setObjectName("labelWindow")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.labelWindow)
        self.comboBoxWindow = QtWidgets.QComboBox(self.groupBoxSmoothing)
        self.comboBoxWindow.setObjectName("comboBoxWindow")
        self.comboBoxWindow.addItem("")
        self.comboBoxWindow.setItemText(0, "")
        self.comboBoxWindow.addItem("")
        self.comboBoxWindow.addItem("")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.comboBoxWindow)
        self.gridLayout_2.addWidget(self.groupBoxSmoothing, 1, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.pushButtonCancel = QtWidgets.QPushButton(singleChannelDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonApply = QtWidgets.QPushButton(singleChannelDialog)
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.horizontalLayout.addWidget(self.pushButtonApply)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(singleChannelDialog)
        self.pushButtonCancel.clicked.connect(singleChannelDialog.reject)
        self.pushButtonApply.clicked.connect(singleChannelDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(singleChannelDialog)

    def retranslateUi(self, singleChannelDialog):
        _translate = QtCore.QCoreApplication.translate
        singleChannelDialog.setWindowTitle(_translate("singleChannelDialog", "Meggie - Single channel"))
        self.groupBoxLegend.setTitle(_translate("singleChannelDialog", "Legend:"))
        self.groupBoxValueRange.setTitle(_translate("singleChannelDialog", "Value range:"))
        self.labelMax.setText(_translate("singleChannelDialog", "Max value:"))
        self.labelMin.setText(_translate("singleChannelDialog", "Min value:"))
        self.groupBoxGeneral.setTitle(_translate("singleChannelDialog", "General:"))
        self.labelChannel.setText(_translate("singleChannelDialog", "Channel:"))
        self.labelTitle.setText(_translate("singleChannelDialog", "Title:"))
        self.groupBoxSmoothing.setTitle(_translate("singleChannelDialog", "Smoothing settings:"))
        self.labelWindowLength.setText(_translate("singleChannelDialog", "Window length:"))
        self.labelWindow.setText(_translate("singleChannelDialog", "Window:"))
        self.comboBoxWindow.setItemText(1, _translate("singleChannelDialog", "hanning"))
        self.comboBoxWindow.setItemText(2, _translate("singleChannelDialog", "flat"))
        self.pushButtonCancel.setText(_translate("singleChannelDialog", "Cancel"))
        self.pushButtonApply.setText(_translate("singleChannelDialog", "Apply"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'aboutDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(921, 591)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.pushButtonClose = QtWidgets.QPushButton(Dialog)
        self.pushButtonClose.setObjectName("pushButtonClose")
        self.horizontalLayout_2.addWidget(self.pushButtonClose)
        self.gridLayout.addLayout(self.horizontalLayout_2, 2, 0, 1, 1)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelVersion = QtWidgets.QLabel(Dialog)
        self.labelVersion.setObjectName("labelVersion")
        self.horizontalLayout.addWidget(self.labelVersion)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.lineEditVersion = QtWidgets.QLineEdit(Dialog)
        self.lineEditVersion.setMaxLength(100)
        self.lineEditVersion.setReadOnly(True)
        self.lineEditVersion.setObjectName("lineEditVersion")
        self.horizontalLayout.addWidget(self.lineEditVersion)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.textBrowser = QtWidgets.QTextBrowser(Dialog)
        self.textBrowser.setObjectName("textBrowser")
        self.verticalLayout.addWidget(self.textBrowser)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Meggie - About"))
        self.pushButtonClose.setText(_translate("Dialog", "Close"))
        self.labelVersion.setText(_translate("Dialog", "Current version:"))
        self.textBrowser.setHtml(_translate("Dialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans Serif\'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'MS Shell Dlg 2\'; font-size:11pt;\">Meggie is realeased under the Simplified BSD lisence:</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'MS Shell Dlg 2\'; font-size:11pt;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">Copyright (c) &lt;2021&gt;, &lt;Kari Aliranta, Jaakko Leppäkangas, Janne Pesonen, Atte Rautio and CIBR&gt;</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">All rights reserved.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">Redistribution and use in source and binary forms, with or without</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">modification, are permitted provided that the following conditions are met: </span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">1. Redistributions of source code must retain the above copyright notice, this</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">   list of conditions and the following disclaimer. </span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">2. Redistributions in binary form must reproduce the above copyright notice,</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">   this list of conditions and the following disclaimer in the documentation</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">   and/or other materials provided with the distribution. </span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS &quot;AS IS&quot; AND</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">The views and conclusions contained in the software and documentation are those</span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">of the authors and should not be interpreted as representing official policies, </span></p>\n"
"<p style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background-color:#f9f9f9;\"><span style=\" font-family:\'monospace,Courier\'; font-size:11pt; color:#000000; background-color:#f9f9f9;\">either expressed or implied, of the FreeBSD Project.</span></p></body></html>"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'activePluginsDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_activePluginsDialog(object):
    def setupUi(self, activePluginsDialog):
        activePluginsDialog.setObjectName("activePluginsDialog")
        activePluginsDialog.resize(414, 176)
        self.gridLayout = QtWidgets.QGridLayout(activePluginsDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.listWidgetPlugins = QtWidgets.QListWidget(activePluginsDialog)
        self.listWidgetPlugins.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)
        self.listWidgetPlugins.setDefaultDropAction(QtCore.Qt.IgnoreAction)
        self.listWidgetPlugins.setSelectionMode(QtWidgets.QAbstractItemView.MultiSelection)
        self.listWidgetPlugins.setObjectName("listWidgetPlugins")
        self.gridLayout.addWidget(self.listWidgetPlugins, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pushButtonCancel = QtWidgets.QPushButton(activePluginsDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonAccept = QtWidgets.QPushButton(activePluginsDialog)
        self.pushButtonAccept.setObjectName("pushButtonAccept")
        self.horizontalLayout.addWidget(self.pushButtonAccept)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(activePluginsDialog)
        self.pushButtonAccept.clicked.connect(activePluginsDialog.accept)
        self.pushButtonCancel.clicked.connect(activePluginsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(activePluginsDialog)
        activePluginsDialog.setTabOrder(self.listWidgetPlugins, self.pushButtonCancel)
        activePluginsDialog.setTabOrder(self.pushButtonCancel, self.pushButtonAccept)

    def retranslateUi(self, activePluginsDialog):
        _translate = QtCore.QCoreApplication.translate
        activePluginsDialog.setWindowTitle(_translate("activePluginsDialog", "Meggie - Active plugins"))
        self.pushButtonCancel.setText(_translate("activePluginsDialog", "Cancel"))
        self.pushButtonAccept.setText(_translate("activePluginsDialog", "Accept"))



""" Contains functions for plot spectrum action
"""

import mne

import numpy as np
import matplotlib.pyplot as plt

from meggie.utilities.plotting import color_cycle
from meggie.utilities.plotting import create_channel_average_plot
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.channels import iterate_topography
from meggie.utilities.units import get_power_unit


def plot_spectrum_averages(subject, channel_groups, name, log_transformed=True):
    """ Plots spectrum averages.
    """

    subject_name = subject.name

    spectrum = subject.spectrum.get(name)

    data = spectrum.content
    freqs = spectrum.freqs
    ch_names = spectrum.ch_names 
    info = spectrum.info

    colors = color_cycle(len(data))
    conditions = spectrum.content.keys()

    averages = {}
    for key, psd in sorted(data.items()):

        data_labels, averaged_data = average_to_channel_groups(
            psd, info, ch_names, channel_groups)

        for label_idx, label in enumerate(data_labels):
            if not label in averages:
                averages[label] = []
            averages[label].append((key, averaged_data[label_idx]))

    ch_types = sorted(set([label[0] for label in averages.keys()]))
    for ch_type in ch_types:

        ch_groups = sorted([label[1] for label in averages.keys()
                            if label[0] == ch_type])

        def plot_fun(ax_idx, ax):
            ch_group = ch_groups[ax_idx]
            ax.set_title(ch_group)
            ax.set_xlabel('Frequency (Hz)')
            ax.set_ylabel('Power ({})'.format(
                get_power_unit(ch_type, log_transformed)))

            for color_idx, (key, curve) in enumerate(averages[(ch_type, ch_group)]):
                if log_transformed:
                    curve = 10 * np.log10(curve)
                ax.plot(freqs, curve, color=colors[color_idx])

        title = ' '.join([name, ch_type])
        legend = list(zip(conditions, colors))
        create_channel_average_plot(len(ch_groups), plot_fun, title, legend)

    plt.show()


def plot_spectrum_topo(subject, name, log_transformed=True, ch_type='meg'):
    """ Plots spectrum topography.
    """

    subject_name = subject.name
    spectrum = subject.spectrum.get(name)
    data = spectrum.content
    freqs = spectrum.freqs
    ch_names = spectrum.ch_names
    info = spectrum.info
    if ch_type == 'meg':
        picked_channels = [ch_name for ch_idx, ch_name in enumerate(info['ch_names'])
                           if ch_idx in mne.pick_types(info, meg=True, eeg=False)]
    else:
        picked_channels = [ch_name for ch_idx, ch_name in enumerate(info['ch_names'])
                           if ch_idx in mne.pick_types(info, eeg=True, meg=False)]
    info = info.copy().pick_channels(picked_channels)

    colors = color_cycle(len(data))

    def individual_plot(ax, info_idx, names_idx):
        """
        """
        ch_name = ch_names[names_idx]
        for color_idx, (key, psd) in enumerate(sorted(data.items())):

            if log_transformed:
                curve = 10 * np.log10(psd[names_idx])
            else:
                curve = psd[names_idx]

            ax.plot(freqs, curve, color=colors[color_idx],
                    label=key)

        title = ' '.join([name, ch_name])
        ax.figure.canvas.set_window_title(title.replace(' ', '_'))
        ax.figure.suptitle(title)
        ax.set_title('')

        ax.legend()
        ax.set_xlabel('Frequency (Hz)')
        ax.set_ylabel('Power ({})'.format(get_power_unit(
            mne.io.pick.channel_type(info, info_idx),
            log_transformed
        )))

        plt.show()

    fig = plt.figure()

    for ax, info_idx, names_idx in iterate_topography(
            fig, info, ch_names, individual_plot):

        handles = []
        for color_idx, (key, psd) in enumerate(sorted(data.items())):

            if log_transformed:
                curve = 10 * np.log10(psd[names_idx])
            else:
                curve = psd[names_idx]

            handles.append(ax.plot(curve, color=colors[color_idx],
                                   linewidth=0.5, label=key)[0])

    if not handles:
        return

    fig.legend(handles=handles)
    title = '{0}_{1}'.format(name, ch_type)
    fig.canvas.set_window_title(title)
    plt.show()




""" Contains controlling logic for the ICA.
"""

import logging

from copy import deepcopy

import numpy as np
import mne

from meggie.utilities.compare import compare_raws


def compute_ica(raw, n_components, method, max_iter, random_state):
    """ Computes ICA using MNE implementation.
    """

    ica = mne.preprocessing.ICA(
        n_components=n_components,
        method=method,
        max_iter=max_iter,
        random_state=random_state)

    ica.fit(raw)
    return ica


def plot_topographies(ica, n_components):
    """ Plots topographies from the ICA solution.
    """

    figs = ica.plot_components(title='')
    for fig in figs:
        fig.canvas.set_window_title('ICA topographic maps')

    def update_topography_texts():
        """ Change texts in the axes to match names in the dialog """
        idx = 0
        for fig in figs:
            for ax in fig.get_axes():
                if idx > n_components:
                    return

                ax.set_title('Component ' + str(idx), fontsize=12)
                idx += 1

    update_topography_texts()


def plot_sources(raw, ica):
    """ Plots sources of the ica solution.
    """
    sources = ica.get_sources(raw)
    sources.plot(title='ICA time courses')


def plot_properties(raw, ica, picks):
    """ Plots properties for specific ICA components.
    """
    figs = ica.plot_properties(
        raw, picks)
    for fig in figs:
        fig.canvas.set_window_title('ICA properties')

    # fix the names
    idx = 0
    for fig in figs:
        for ax_idx, ax in enumerate(fig.get_axes()):
            if ax_idx == 0:
                ax.set_title("Component " + str(picks[idx]))
                idx += 1
            break


def plot_changes(raw, ica, indices):
    """ Plot a raw comparison plot for ICA solution.
    """
    raw_removed = raw.copy()
    ica.apply(raw_removed, exclude=indices)
    compare_raws(raw, raw_removed)




""" Contains create epochs action handling.
"""

from meggie.utilities.names import next_available_name
from meggie.utilities.threading import threaded

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.epochs_create.dialogs.createEpochsFromEventsDialogMain import CreateEpochsFromEventsDialog
from meggie.actions.epochs_create.controller.epoching import create_epochs_from_events


class CreateEpochs(Action):
    """ Shows a dialog for event selection and then creates epochs.
    """
    def run(self):
        default_name = next_available_name(
            self.experiment.active_subject.spectrum.keys(), 'Epochs')

        dialog = CreateEpochsFromEventsDialog(self.experiment, self.window, 
                                              default_name, self.handler)
        dialog.show()


    @subject_action
    def handler(self, subject, params):
        """
        """
        @threaded
        def create():
            create_epochs_from_events(subject, params)

        create(do_meanwhile=self.window.update_ui)



""" Contains a class for logic of the power spectrum creation dialog.
"""

import logging

from collections import OrderedDict

from PyQt5 import QtWidgets

import numpy as np

from meggie.utilities.dialogs.powerSpectrumDialogUi import Ui_PowerSpectrumDialog
from meggie.utilities.dialogs.powerSpectrumAddAdvancedDialogMain import PowerSpectrumAddAdvancedDialog

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.utilities.validators import validate_name
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox


class PowerSpectrumDialog(QtWidgets.QDialog):
    """ Contains logic for the power spectrum creation dialog.
    """
    def __init__(self, experiment, parent, default_name, handler=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_PowerSpectrumDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment

        self.handler = handler

        self.intervals = []

        raw = self.experiment.active_subject.get_raw()

        tmax = np.floor(raw.times[raw.n_times - 1]) - 0.1
        self.ui.doubleSpinBoxTmin.setValue(0)
        self.ui.doubleSpinBoxTmax.setValue(tmax)
        self.ui.doubleSpinBoxTmin.setMaximum(tmax)
        self.ui.doubleSpinBoxTmax.setMaximum(tmax)

        # set nfft initially to ~2 seconds and overlap to ~1 seconds
        sfreq = raw.info['sfreq']
        window_in_seconds = 2
        nfft = int(
            np.power(2, np.ceil(np.log(sfreq * window_in_seconds) / np.log(2))))
        overlap = nfft / 2

        self.ui.spinBoxNfft.setValue(nfft)
        self.ui.spinBoxOverlap.setValue(overlap)

        if raw.info.get('highpass'):
            if self.ui.spinBoxFmin.value() < raw.info['highpass']:
                self.ui.spinBoxFmin.setValue(
                    int(np.ceil(raw.info['highpass'])))

        if raw.info.get('lowpass'):
            if self.ui.spinBoxFmax.value() > raw.info['lowpass']:
                self.ui.spinBoxFmax.setValue(int(raw.info['lowpass']))

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

        self.ui.lineEditName.setText(default_name)

    def _experiment_getter(self):
        return self.experiment

    def on_pushButtonAdd_clicked(self, checked=None):
        if checked is None:
            return
        group = str(self.ui.comboBoxAvgGroup.currentText())
        tmin = self.ui.doubleSpinBoxTmin.value()
        tmax = self.ui.doubleSpinBoxTmax.value()
        if tmin >= tmax:
            messagebox(
                self.parent, "End time must be higher than the starting time")
            return

        self.add_intervals([('fixed', (group, tmin, tmax))])

    def add_intervals(self, intervals):
        """ Add intervals to the interval list.
        """
        for ival_type, interval in intervals:
            self.intervals.append((ival_type, interval))
            if ival_type == 'fixed':
                item = QtWidgets.QListWidgetItem(
                    '%s: %s - %s s (fixed)' % (
                        interval[0],
                        round(interval[1], 4),
                        round(interval[2], 4)
                    ))
            else:
                item_string = str(interval[0]) + ': ['
                if interval[1][0] == 'events':
                    item_string += ', '.join([str(elem)
                                              for elem in interval[1][1:]])
                elif interval[1][0] == 'start':
                    item_string += 'start, ' + str(interval[1][3])
                elif interval[1][0] == 'end':
                    item_string += 'end, ' + str(interval[1][3])

                item_string += '] – ['

                if interval[2][0] == 'events':
                    item_string += ', '.join([str(elem)
                                              for elem in interval[2][1:]])
                elif interval[2][0] == 'start':
                    item_string += 'start, ' + str(interval[2][3])

                elif interval[2][0] == 'end':
                    item_string += 'end, ' + str(interval[2][3])

                item_string += '] (dynamic)'

                item = QtWidgets.QListWidgetItem(
                    item_string)
            self.ui.listWidgetIntervals.addItem(item)

    def on_pushButtonClear_clicked(self, checked=None):
        if checked is None:
            return
        self.intervals = []
        self.ui.listWidgetIntervals.clear()

    def on_pushButtonClearRow_clicked(self, checked=None):
        if checked is None:
            return

        current_row = self.ui.listWidgetIntervals.currentRow()
        if current_row == -1:
            return

        self.ui.listWidgetIntervals.takeItem(current_row)
        self.intervals.pop(current_row)

    def on_pushButtonAddAdvanced_clicked(self, checked=None):
        if checked is None:
            return
        dialog = PowerSpectrumAddAdvancedDialog(self)
        dialog.show()

    def accept(self, *args, **kwargs):
        try:
            spectrum_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        intervals = self.intervals
        if not intervals:
            return

        fmin = self.ui.spinBoxFmin.value()
        fmax = self.ui.spinBoxFmax.value()

        subject = self.experiment.active_subject

        params = dict()
        params['fmin'] = fmin
        params['fmax'] = fmax
        params['nfft'] = self.ui.spinBoxNfft.value()
        params['overlap'] = self.ui.spinBoxOverlap.value()
        params['intervals'] = intervals
        params['name'] = spectrum_name

        try:
            self.handler(subject, params)
            self.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self, *args):
        try:
            spectrum_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        intervals = self.intervals
        if not intervals:
            return

        fmin = self.ui.spinBoxFmin.value()
        fmax = self.ui.spinBoxFmax.value()

        subject = self.experiment.active_subject

        params = dict()
        params['fmin'] = fmin
        params['fmax'] = fmax
        params['nfft'] = self.ui.spinBoxNfft.value()
        params['overlap'] = self.ui.spinBoxOverlap.value()
        params['intervals'] = intervals
        params['name'] = spectrum_name

        selected_subject_names = self.batching_widget.selected_subjects
        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append((subject,
                                                                 str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        try:
            self.experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()



""" Contains save tse action handling.
"""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.names import next_available_name
from meggie.utilities.channels import get_channels_by_type
from meggie.utilities.validators import assert_arrays_same

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.utilities.dialogs.TFROutputOptionsMain import TFROutputOptions

from meggie.actions.tfr_save_tse.controller.tfr import save_tse_channel_averages
from meggie.actions.tfr_save_tse.controller.tfr import save_tse_all_channels


class SaveTSE(Action):
    """ Saves TSE's to csv files """

    def run(self):
        try:
            selected_name = self.data['outputs']['tfr'][0]
        except IndexError as exc:
            return

        time_arrays = []
        freq_arrays = []
        for subject in self.experiment.subjects.values():
            tfr = subject.tfr.get(selected_name)
            if not tfr:
                continue
            time_arrays.append(tfr.times)
            freq_arrays.append(tfr.freqs)
        assert_arrays_same(time_arrays)
        assert_arrays_same(freq_arrays, 'Freqs do no match')

        def option_handler(params):
            params['channel_groups'] = self.experiment.channel_groups
            params['name'] = selected_name

            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = TFROutputOptions(self.window, self.experiment,
                                  selected_name, handler=option_handler)
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """
        """
        if params['output_option'] == 'all_channels':
            save_tse_all_channels(
                self.experiment, params['name'],
                params['blmode'], params['blstart'], params['blend'],
                params['tmin'], params['tmax'], params['fmin'], params['fmax'],
                do_meanwhile=self.window.update_ui)
        else:
            save_tse_channel_averages(
                self.experiment, params['name'],
                params['blmode'], params['blstart'], params['blend'],
                params['tmin'], params['tmax'], params['fmin'], params['fmax'],
                params['channel_groups'], do_meanwhile=self.window.update_ui)




""" Contains controlling logic for tfr create
"""

import mne

from meggie.utilities.threading import threaded
from meggie.utilities.validators import assert_arrays_same

from meggie.datatypes.tfr.tfr import TFR


@threaded
def create_tfr(subject, tfr_name, epochs_names,
               freqs, decim, n_cycles, subtract_evoked):
    """ Handles tfr item creation.
    """

    time_arrays = []
    for name in epochs_names:
        collection = subject.epochs.get(name)
        if collection:
            time_arrays.append(collection.content.times)
    assert_arrays_same(time_arrays)

    tfrs = {}
    for epoch_name in epochs_names:
        epochs = subject.epochs[epoch_name].content
        if subtract_evoked:
            epochs = epochs.copy().subtract_evoked()

        tfr = mne.time_frequency.tfr.tfr_morlet(epochs, 
                                                freqs=freqs, 
                                                n_cycles=n_cycles,
                                                decim=decim, 
                                                average=True,
                                                return_itc=False)
        tfrs[epoch_name] = tfr

    # convert list-like to list
    if hasattr(n_cycles, '__len__'):
        n_cycles = list(n_cycles)

    params = {
        'decim': decim,
        'n_cycles': n_cycles,
        'evoked_subtracted': subtract_evoked,
        'conditions': epochs_names
    }

    meggie_tfr = TFR(tfr_name, subject.tfr_directory, params, tfrs)

    meggie_tfr.save_content()
    subject.add(meggie_tfr, "tfr")




import os

import mne
import matplotlib.pyplot as plt

from meggie.utilities.channels import get_triplet_from_mag
from meggie.utilities.channels import get_channels_by_type
from meggie.utilities.channels import get_default_channel_groups
from meggie.utilities.channels import clean_names
from meggie.utilities.channels import iterate_topography
from meggie.utilities.channels import average_to_channel_groups


def test_get_triplet_from_mag():
    assert(get_triplet_from_mag('MEG2031') == ['MEG2031', 'MEG2032', 'MEG2033'])


def test_get_channels_by_type():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, 'MEG', 'sample', 'sample_audvis_raw.fif')
    info = mne.io.read_raw_fif(sample_fname).info

    chs = get_channels_by_type(info)
    assert(set(chs.keys()) == set(['grad', 'mag', 'eeg']))
    assert(len(chs['eeg']) == 59)
    assert(len(chs['grad']) == 203)
    assert(len(chs['mag']) == 102)


def test_default_channel_groups():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, 'MEG', 'sample', 'sample_audvis_raw.fif')
    info = mne.io.read_raw_fif(sample_fname).info

    assert(get_default_channel_groups(info, 'eeg')['Left-frontal'] ==
           ['EEG 001', 'EEG 004', 'EEG 005', 'EEG 009', 'EEG 010', 'EEG 011', 'EEG 012'])


def test_clean_names():
    assert(clean_names(['EEG 012']) == clean_names(['EEG012']) == ['EEG012'])


def test_iterate_topography():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, 'MEG', 'sample', 'sample_audvis_raw.fif')
    info = mne.io.read_raw_fif(sample_fname).info

    fig = plt.figure()

    ch_names = info['ch_names'][::2]

    def on_pick(ax, info_idx, names_idx):
        pass

    iterated = []
    for ax, info_idx, names_idx in iterate_topography(fig, info, ch_names, on_pick):
        iterated.append((info_idx, names_idx))

    assert(iterated[0] == (0,0))
    assert(iterated[-1] == (304, 152))


def test_average_to_channel_groups():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, 'MEG', 'sample', 'sample_audvis_raw.fif')
    raw = mne.io.read_raw_fif(sample_fname, preload=True)
    info = raw.info

    data = raw._data

    ch_names = info['ch_names'][:20]

    meg_channel_groups = get_default_channel_groups(info, 'meg')
    eeg_channel_groups = get_default_channel_groups(info, 'eeg')

    # find out to which channel group each of the channels belongs to
    results = []
    for ch_name in ch_names:
        for ch_group_name, ch_group in meg_channel_groups.items():
            if ch_name in ch_group:
                results.append((ch_name, ch_group_name))

    labels, averaged_data = average_to_channel_groups(raw._data[:, 0:100], info, 
        ch_names, {'meg': meg_channel_groups, 'eeg': eeg_channel_groups})

    # check that the localization of data is same before and after averaging
    assert(set([result[1] for result in results]) == set([label[1] for label in labels]))




""" Contains spectrum creation logic.
"""

from copy import deepcopy
from collections import OrderedDict

import mne

import numpy as np

from meggie.datatypes.spectrum.spectrum import Spectrum

from meggie.utilities.threading import threaded
from meggie.utilities.events import get_raw_blocks_from_intervals


@threaded
def create_power_spectrum(subject, spectrum_name, intervals, 
                          fmin, fmax, nfft, overlap):
    """ Creates a power spectrum item.
    """
    # get raw objects organized with average groups as keys
    ival_times, raw_block_groups = get_raw_blocks_from_intervals(subject,
                                                                 intervals)

    raw = subject.get_raw()
    picks = mne.pick_types(raw.info, meg=True, eeg=True,
                           exclude='bads')

    # remove zero channels from picks
    zero_idxs = []
    for idx, row in enumerate(raw._data):
        if np.all(row == 0):
            zero_idxs.append(idx)
    picks = [pick for pick in picks if pick not in zero_idxs]


    # compute psd's
    psd_groups = OrderedDict()
    for key, raw_blocks in raw_block_groups.items():
        for raw_block in raw_blocks:
            length = len(raw_block.times)
            psds, freqs = mne.time_frequency.psd_welch(
                raw_block, fmin=fmin, fmax=fmax,
                n_fft=nfft, n_overlap=overlap, picks=picks,
                proj=True)

            if key not in psd_groups:
                psd_groups[key] = []

            psd_groups[key].append((psds, freqs, length))

    for psd_list in psd_groups.values():
        freqs = psd_list[0][1]
        break

    psds = []
    for psd_list in psd_groups.values():
        # do a weighted (raw block lengths as weights) average of psds inside a
        # group
        weights = np.array([length for psds_, freqs, length in psd_list])
        weights = weights.astype(float) / np.sum(weights)
        psd = np.average([psds_ for psds_, freqs, length in psd_list],
                         weights=weights, axis=0)
        psds.append(psd)

    info = mne.pick_info(raw.info, sel=picks)
    psd_data = dict(zip(psd_groups.keys(), psds))

    params = {}
    params['fmin'] = fmin
    params['fmax'] = fmax
    params['nfft'] = nfft
    params['overlap'] = overlap
    params['conditions'] = [elem for elem in psd_groups.keys()]
    params['intervals'] = ival_times

    spectrum = Spectrum(spectrum_name, subject.spectrum_directory,
                        params, psd_data, freqs, info)

    spectrum.save_content()
    subject.add(spectrum, 'spectrum')




""" Provides functions for popping up different kind of messageboxes.
"""

import logging
from meggie.utilities.dialogs.shortMessageBoxMain import shortMessageBox
from meggie.utilities.dialogs.shortQuestionBoxMain import shortQuestionBox


def exc_messagebox(parent, exc, exec_=False):
    """Pops up a messagebox for a exception.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    exc : instance of Exception
        The exception that is presented.
    exec_ : bool
        Whether to make the messagebox modal.
    """

    try:
        if isinstance(exc, str):
            error_message = exc
        else:
            error_message = str(exc)
            if not error_message:
                error_message = ('(empty message of type ' +
                                 str(exc.__repr__()) + ')')
    except Exception as e:
        error_message = ''

    # print traceback to console
    logging.getLogger('ui_logger').exception('')

    # create messagebox for user
    message = '\n\n'.join([
        'There has been an error with following message: ',
        error_message,
        'More information can be found from console below.',
    ])

    messagebox = shortMessageBox(message, parent)
    if exec_:
        messagebox.exec_()
    else:
        messagebox.show()


def messagebox(parent, msg, exec_=False):
    """ Pops up a messagebox with a message.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    msg : str
        The message that is presented.
    exec_ : bool
        Whether to make the messagebox modal.

    """
    messagebox = shortMessageBox(msg, parent)
    if exec_:
        messagebox.exec_()
    else:
        messagebox.show()


def questionbox(parent, question, handler):
    """ Pops up a yes/no questionbox.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    question : str
        The question that is asked.
    handler : function
        A function that is called with the answer used provided.

    """

    questionbox = shortQuestionBox(question, parent, handler)
    questionbox.exec_()



""" Contains a class for logic of the output options dialog.
"""
import logging

from PyQt5 import QtWidgets

from meggie.utilities.dialogs.outputOptionsUi import Ui_outputOptions


class OutputOptions(QtWidgets.QDialog):
    """ Contains logic for the output options dialog.
    """

    def __init__(self, parent, handler=None,
                 selected_option=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_outputOptions()
        self.ui.setupUi(self)
        self.parent = parent
        self.handler = handler

        if selected_option == 'channel_averages':
            self.ui.radioButtonChannelAverages.setChecked(True)

    def accept(self):
        if self.ui.radioButtonChannelAverages.isChecked():
            selected_option = 'channel_averages'
        else:
            selected_option = 'all_channels'

        if self.handler:
            self.handler(selected_option=selected_option)

        self.close()



""" Contains a class for logic of filter dialog.
"""

import logging

from copy import deepcopy

from PyQt5 import QtCore
from PyQt5 import QtWidgets

from meggie.actions.raw_filter.dialogs.filterDialogUi import Ui_DialogFilter
from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.actions.raw_filter.controller.filter import filter_data

from meggie.utilities.compare import compare_raws
from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox


class FilterDialog(QtWidgets.QDialog):
    """ Contains logic for filter dialog.
    """

    def __init__(self, parent, experiment, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_DialogFilter()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment
        self.handler = handler

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

    def _experiment_getter(self):
        return self.experiment

    def on_pushButtonPreview_clicked(self, checked=None):
        if checked is None:
            return

        params = self._collect_parameter_values()
        if not params:
            message = 'No filter(s) selected.'
            messagebox(self.parent, message)
            return

        subject = self.experiment.active_subject

        try:
            raw_to = filter_data(subject,
                                 params,
                                 preview=True)
            raw_from = subject.get_raw()
            compare_raws(raw_from, raw_to)
        except Exception as exc:
            exc_messagebox(self.parent, exc)

    def accept(self):
        subject = self.experiment.active_subject
        params = self._collect_parameter_values()
        if not params:
            message = 'No filter(s) selected'
            messagebox(self.parent, message)
            return

        try:
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):
        subject_names = self.batching_widget.selected_subjects
        params = self._collect_parameter_values()
        if not params:
            message = 'No filter(s) selected'
            messagebox(self.parent, message)
            return

        for name, subject in self.experiment.subjects.items():
            if name in subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    logging.getLogger('ui_logger').exception('')
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))

        self.batching_widget.cleanup()
        self.parent.initialize_ui()
        self.close()

    def _collect_parameter_values(self):
        is_empty = True
        dictionary = {}

        length = str(self.ui.doubleSpinBoxLength.value()) + 's'
        dictionary['length'] = length

        dictionary['trans_bw'] = self.ui.doubleSpinBoxTransBandwidth.value()

        dictionary['lowpass'] = self.ui.checkBoxLowpass.isChecked()
        if dictionary['lowpass']:
            is_empty = False
            dictionary['low_cutoff_freq'] = \
                self.ui.doubleSpinBoxLowpassCutoff.value()

        dictionary['highpass'] = self.ui.checkBoxHighpass.isChecked()
        if dictionary['highpass']:
            is_empty = False
            dictionary['high_cutoff_freq'] = \
                self.ui.doubleSpinBoxHighpassCutoff.value()

        dictionary['bandstop1'] = self.ui.checkBoxBandstop.isChecked()
        if dictionary['bandstop1']:
            is_empty = False
            dictionary['bandstop1_freq'] = \
                self.ui.doubleSpinBoxBandstopFreq.value()

        dictionary['bandstop2'] = self.ui.checkBoxBandstop2.isChecked()
        if dictionary['bandstop2']:
            is_empty = False
            dictionary['bandstop2_freq'] = \
                self.ui.doubleSpinBoxBandstopFreq2.value()

        dictionary['bandstop_bw'] = self.ui.doubleSpinBoxBandstopWidth.value()
        dictionary['bandstop_transbw'] = self.ui.doubleSpinBoxNotchTransBw.value()  # noqa

        length = str(self.ui.doubleSpinBoxBandStopLength.value()) + 's'
        dictionary['bandstop_length'] = length

        return dictionary




""" Contains a class for logic of action dialog.
"""
import os
import json
import logging
import re

from pprint import pformat

from PyQt5 import QtWidgets

from meggie.mainwindow.dynamic import find_all_action_specs
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox

from meggie.mainwindow.dialogs.actionDialogUi import Ui_ActionDialog


class ActionDialog(QtWidgets.QDialog):
    """ Contains logic for action dialog.
    """

    def __init__(self, parent, experiment):
        QtWidgets.QDialog.__init__(self, parent)
        self.parent = parent
        self.ui = Ui_ActionDialog()
        self.ui.setupUi(self)

        # clear the tree widget
        self.ui.treeWidgetActions.clear()

        # read the actions log
        try:
            path = os.path.join(experiment.path, 'actions.log')
            with open(path, 'r') as f:
                lines = f.readlines()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        # get action specifications for id -> name conversion
        action_specs = find_all_action_specs()

        # organize actions data by action uid
        actions = {}
        for line in lines:
            content = json.loads(line)
            if not content['uid'] in actions:
                actions[content['uid']] = []
            actions[content['uid']].append(content)

        # combine data of each action uid to single dict
        # and organize by subject uid
        subject_tree = {}
        for action_uid, action_rows in actions.items():
            action_tree = {}
            name = None
            for row in action_rows:
                if row['type'] == 'ACTION_START':
                    id_ = row['id']
                    desc = row.get('desc')

                if row['type'] == 'SUBJECT_START':
                    if not row['subject_uid'] in action_tree:
                        action_tree[row['subject_uid']] = {}

                    action_tree[row['subject_uid']]['params'] = row['params']

                if row['type'] == 'SUBJECT_END':
                    if not row['subject_uid'] in action_tree:
                        action_tree[row['subject_uid']] = {}
                    action_tree[row['subject_uid']]['finished'] = True
                    action_tree[row['subject_uid']]['timestamp'] = row['timestamp']

            for subject_uid, branch in action_tree.items():
                if branch.get('finished'):
                    branch['id'] = id_
                    branch['desc'] = desc
                    if subject_uid not in subject_tree:
                        subject_tree[subject_uid] = []
                    subject_tree[subject_uid].append(branch)

        # for each subject of the experiment (subjects that have been removed
        # or replaced will not be showed) see if there are entries of finished
        # actions and add them to the tree widget.
        for subject_name, subject in sorted(experiment.subjects.items()):
            if subject.uid not in subject_tree:
                continue

            # Add root node for subject
            subject_item = QtWidgets.QTreeWidgetItem(self.ui.treeWidgetActions)
            subject_item.setText(0, subject_name)

            subject_actions = subject_tree[subject.uid]
            for action_data in subject_actions:

                # create node for each action
                action_item = QtWidgets.QTreeWidgetItem(subject_item)
                
                # if name can be found from the core or installed plugins,
                # use it, otherwise print just the id.
                name = action_data['id']
                if action_data['id'] in action_specs:
                    if action_specs[action_data['id']][2].get('name'):
                        name = action_specs[action_data['id']][2].get('name')

                if action_data['desc']:
                    name += " ({0})".format(action_data['desc'])

                action_item.setText(0, name)

                # Add params item to the action
                params_item = QtWidgets.QTreeWidgetItem(action_item)
                params_item.setText(0, 'Params')
  
                # And fill it with params that are present in the data
                for param_name, param_value in action_data['params'].items():
                    msg = "{0}: {1}".format(param_name, json.dumps(param_value))
                    param_item = QtWidgets.QTreeWidgetItem(params_item)
                    param_item.setText(0, msg)

                # Get date and utc time from the data
                template = re.compile(r'([0-9]*-[0-9]*-[0-9]*)T([0-9]*:[0-9]*:[0-9]*)\..*')
                elems = template.match(action_data['timestamp'])
                msg = "{0}: {1} {2} UTC".format("Timestamp", elems[1], elems[2])
       
                # and add a item
                timestamp_item = QtWidgets.QTreeWidgetItem(action_item)
                timestamp_item.setText(0, msg)

    def on_treeWidgetActions_itemDoubleClicked(self, item):
        if item.childCount() == 0:
            # try creating a pretty formatted details dialog
            try:
                match = re.compile("([a-zA-Z0-9_]*): (.*)").match(item.text(0))
                if not match:
                    return

                key = match.group(1)
                val = match.group(2)
                try:
                    obj = json.loads(val)
                except Exception as exc:
                    obj = val
                message = pformat(obj)
                messagebox(self, message)
            except Exception as exc:
                logging.getLogger('ui_logger').exception("")
                return




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'outputOptionsUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_outputOptions(object):
    def setupUi(self, outputOptions):
        outputOptions.setObjectName("outputOptions")
        outputOptions.resize(255, 149)
        self.gridLayout = QtWidgets.QGridLayout(outputOptions)
        self.gridLayout.setObjectName("gridLayout")
        self.groupBoxType = QtWidgets.QGroupBox(outputOptions)
        self.groupBoxType.setObjectName("groupBoxType")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.groupBoxType)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.radioButtonAllChannels = QtWidgets.QRadioButton(self.groupBoxType)
        self.radioButtonAllChannels.setChecked(True)
        self.radioButtonAllChannels.setObjectName("radioButtonAllChannels")
        self.gridLayout_2.addWidget(self.radioButtonAllChannels, 0, 0, 1, 1)
        self.radioButtonChannelAverages = QtWidgets.QRadioButton(self.groupBoxType)
        self.radioButtonChannelAverages.setObjectName("radioButtonChannelAverages")
        self.gridLayout_2.addWidget(self.radioButtonChannelAverages, 1, 0, 1, 1)
        self.gridLayout.addWidget(self.groupBoxType, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pushButtonCancel = QtWidgets.QPushButton(outputOptions)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonAccept = QtWidgets.QPushButton(outputOptions)
        self.pushButtonAccept.setObjectName("pushButtonAccept")
        self.horizontalLayout.addWidget(self.pushButtonAccept)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem1, 1, 0, 1, 1)

        self.retranslateUi(outputOptions)
        self.pushButtonCancel.clicked.connect(outputOptions.reject)
        self.pushButtonAccept.clicked.connect(outputOptions.accept)
        QtCore.QMetaObject.connectSlotsByName(outputOptions)

    def retranslateUi(self, outputOptions):
        _translate = QtCore.QCoreApplication.translate
        outputOptions.setWindowTitle(_translate("outputOptions", "Meggie - Output options"))
        self.groupBoxType.setTitle(_translate("outputOptions", "Type"))
        self.radioButtonAllChannels.setText(_translate("outputOptions", "All channels"))
        self.radioButtonChannelAverages.setText(_translate("outputOptions", "Channel averages"))
        self.pushButtonCancel.setText(_translate("outputOptions", "Cancel"))
        self.pushButtonAccept.setText(_translate("outputOptions", "Accept"))




""" Contains functions for name-related tasks.
"""

import re


def next_available_name(old_names, stem):
    """Given list of existing names (such as cat_1, cat_2) 
    and a stem (cat), find next name in order (cat_3).

    Used for default names in the dialogs.

    Parameters
    ----------
    old_names : list
        Already existing names.
    stem : str
        Stem of all names.

    Returns
    -------
    str
        The next available name.

    """
    suffices = []
    for old_name in old_names:
        template = stem + r'([_]?)' + r'([0-9]*)$'
        match = re.match(template, old_name)
        if match:
            suffix = match.group(2)
            if suffix == '':
                suffix = 0
            suffices.append(int(suffix))

    if not suffices:
        return stem

    return stem + '_' + str(max(suffices)+1)




""" Contains controlling logic for the epochs creation.
"""

import copy
import logging

import numpy as np
import mne

from meggie.utilities.units import get_scaling
from meggie.utilities.events import find_stim_channel
from meggie.utilities.events import find_events

from meggie.datatypes.epochs.epochs import Epochs


def create_epochs_from_events(subject, params):
    """ Creates epochs based on events.
    """
    raw = subject.get_raw()

    params = copy.deepcopy(params)
    event_params = params['events']
    reject_params = params['reject']

    # convert reject params from human readable units to standard units
    for key in reject_params:
        reject_params[key] /= get_scaling(key)

    # remove params that don't match with the channel types present in raw
    if mne.pick_types(raw.info, meg='grad', eeg=False).size == 0:
        reject_params.pop('grad', None)
    if mne.pick_types(raw.info, meg='mag', eeg=False).size == 0:
        reject_params.pop('mag', None)
    if mne.pick_types(raw.info, meg=False, eeg=True).size == 0:
        reject_params.pop('eeg', None)

    events = []
    category = {}

    stim_channel = find_stim_channel(raw)

    # event_id should not matter after epochs are created,
    # so we just add placeholders as 1, 2, 3...
    if len(event_params) > 0:
        for idx, item in enumerate(event_params):
            event_id = item['event_id']
            mask = item['mask']
            category_id = (
                'id_' + str(event_id) + '_mask_' + str(mask))

            new_events = find_events(raw, stim_channel, mask, event_id)

            if len(new_events) == 0:
                logging.warning('No events found with setting ' +
                                str(category_id))
                continue

            category[category_id] = idx + 1
            new_events[:, 2] = idx + 1

            events.extend([[event[0] + int(round(raw.info['sfreq']*params['delay'])), 
                            event[1], event[2]] 
                           for event in new_events])

    if len(events) == 0:
        raise ValueError(
            'No matching events found. Please check rejection limits and other parameters.')

    # prepare parameters for pick_types
    if params['mag'] and params['grad']:
        meg = True
    elif params['mag']:
        meg = 'mag'
    elif params['grad']:
        meg = 'grad'
    else:
        meg = False

    eeg = params['eeg']

    # find all proper picks, dont exclude bads
    picks = mne.pick_types(raw.info, meg=meg, eeg=eeg, exclude=[])

    if len(picks) == 0:
        raise ValueError('You should select at least one channel type')

    mne_epochs = mne.Epochs(raw, np.array(events),
                            category, params['tmin'], params['tmax'],
                            baseline=(params['bstart'], params['bend']),
                            picks=picks, reject=reject_params)

    if len(mne_epochs.get_data()) == 0:
        raise ValueError('Could not find any data. Perhaps the ' +
                         'rejection thresholds are too strict...')

    n_dropped = len(events) - len(mne_epochs.get_data())

    if n_dropped > 0:
        logging.getLogger('ui_logger').info(str(n_dropped) + ' epochs dropped.')

    epochs_directory = subject.epochs_directory
    epochs = Epochs(params['collection_name'],
                    epochs_directory,
                    params,
                    content=mne_epochs)

    epochs.save_content()
    subject.add(epochs, 'epochs')




""" Contains a class for logic of TFR creation dialog.
"""
import logging

import numpy as np

from PyQt5 import QtWidgets

from meggie.actions.tfr_create.dialogs.TFRDialogUi import Ui_TFRDialog

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.utilities.validators import validate_name
from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox


class TFRDialog(QtWidgets.QDialog):
    """ Contains logic for the TFR creation dialog.
    """

    def __init__(self, experiment, parent, epoch_names, 
                 default_name, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_TFRDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.epoch_names = epoch_names
        self.experiment = experiment
        self.handler = handler

        subject = experiment.active_subject
        epochs = subject.epochs[epoch_names[0]].content

        self.ui.lineEditEpochName.setText(', '.join(epoch_names))

        if epochs.info.get('highpass'):
            self.ui.doubleSpinBoxMinFreq.setValue(
                max(int(np.ceil(epochs.info['highpass'])),
                    self.ui.doubleSpinBoxMinFreq.value()))

        if epochs.info.get('lowpass'):
            self.ui.doubleSpinBoxMaxFreq.setValue(
                min(int(np.floor(epochs.info['lowpass'])),
                    self.ui.doubleSpinBoxMaxFreq.value()))

        epoch_length = epochs.times[-1] - epochs.times[0]

        # try to use as many cycles as possible
        # (window ~= 10 * (n_cycles / (2.0 * np.pi * freq))
        minfreq = self.ui.doubleSpinBoxMinFreq.value()
        n_cycles = epoch_length * 2.0 * np.pi * minfreq / 10.0
        n_cycles = max(np.floor(n_cycles), 1)

        self.ui.doubleSpinBoxNcycles.setValue(n_cycles)

        # select factor such that minfreq / factor = n_cycles,
        # and then ceil
        factor = np.ceil(minfreq / float(n_cycles))
        self.ui.doubleSpinBoxCycleFactor.setValue(factor)

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

        self.ui.lineEditTFRName.setText(default_name)

    def _experiment_getter(self):
        return self.experiment

    def accept(self):
        tfr_name = self.ui.lineEditTFRName.text()

        try:
            validate_name(tfr_name)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        minfreq = self.ui.doubleSpinBoxMinFreq.value()
        maxfreq = self.ui.doubleSpinBoxMaxFreq.value()
        interval = self.ui.doubleSpinBoxFreqInterval.value()
        freqs = np.arange(minfreq, maxfreq, interval)

        decim = self.ui.spinBoxDecim.value()

        subtract_evoked = self.ui.checkBoxSubtractEvoked.isChecked()

        if self.ui.radioButtonFixed.isChecked():
            n_cycles = self.ui.doubleSpinBoxNcycles.value()
        elif self.ui.radioButtonAdapted.isChecked():
            n_cycles = freqs / self.ui.doubleSpinBoxCycleFactor.value()

        params = {}
        params['freqs'] = freqs
        params['decim'] = decim
        params['subtract_evoked'] = subtract_evoked
        params['n_cycles'] = n_cycles
        params['name'] = tfr_name
        params['conditions'] = self.epoch_names

        experiment = self.experiment
        subject = experiment.active_subject

        try:
            self.handler(subject, params)
            experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        self.parent.initialize_ui()

        self.close()

    def acceptBatch(self):
        tfr_name = self.ui.lineEditTFRName.text()

        try:
            validate_name(tfr_name)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        minfreq = self.ui.doubleSpinBoxMinFreq.value()
        maxfreq = self.ui.doubleSpinBoxMaxFreq.value()
        decim = self.ui.spinBoxDecim.value()
        interval = self.ui.doubleSpinBoxFreqInterval.value()
        freqs = np.arange(minfreq, maxfreq, interval)

        subtract_evoked = self.ui.checkBoxSubtractEvoked.isChecked()

        if self.ui.radioButtonFixed.isChecked():
            n_cycles = self.ui.doubleSpinBoxNcycles.value()
        elif self.ui.radioButtonAdapted.isChecked():
            n_cycles = freqs / self.ui.doubleSpinBoxCycleFactor.value()

        params = {}
        params['freqs'] = freqs
        params['decim'] = decim
        params['subtract_evoked'] = subtract_evoked
        params['n_cycles'] = n_cycles
        params['name'] = tfr_name
        params['conditions'] = self.epoch_names

        experiment = self.experiment

        selected_subject_names = self.batching_widget.selected_subjects
        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        try:
            experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self.parent, exc)

        self.parent.initialize_ui()

        self.close()



""" Contains a class for logic of the rereferencing dialog.
"""
import logging

import numpy as np
import mne

from PyQt5 import QtWidgets

from meggie.actions.raw_rereference.dialogs.rereferencingDialogUi import Ui_rereferencingDialog

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget
from meggie.utilities.messaging import exc_messagebox


class RereferencingDialog(QtWidgets.QDialog):
    """ Contains logic for the rereferencing dialog.
    """
    def __init__(self, parent, experiment, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_rereferencingDialog()
        self.ui.setupUi(self)

        self.experiment = experiment
        self.parent = parent
        self.handler = handler

        subject = self.experiment.active_subject
        raw = subject.get_raw()
        sfreq = raw.info['sfreq']

        # fill the combobox
        picks = mne.pick_types(raw.info, eeg=True, meg=False, eog=True)
        ch_names = [ch_name for ch_idx, ch_name in
                    enumerate(raw.info['ch_names']) if ch_idx in picks]

        for ch_name in ch_names:
            self.ui.comboBoxChannel.addItem(ch_name)

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry())
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

    def _experiment_getter(self):
        return self.experiment

    def accept(self):
        experiment = self.experiment
        subject = experiment.active_subject

        selection = self.ui.comboBoxChannel.currentText()

        try:
            params = {'selection': selection}
            self.handler(subject, params)
            experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self.parent, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self):
        experiment = self.experiment
        selection = self.ui.comboBoxChannel.currentText()

        selected_subject_names = self.batching_widget.selected_subjects

        for name, subject in experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    params = {'selection': selection}
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append(
                        (subject, str(exc)))
                    logging.getLogger('ui_logger').exception('')

        self.batching_widget.cleanup()

        try:
            experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self.parent, exc)

        self.parent.initialize_ui()

        self.close()




""" Defines Evoked class, wraps mne.Evoked objects.
"""

import os
import logging

import mne

from meggie.utilities.datatype import Datatype


class Evoked(Datatype):
    """ A wrapper for mne.Evoked objects.

    Parameters
    ----------
    name : str
        Name of the evoked, used in the UI lists and in the .exp file.
    directory : str
        Absolute path to the data folder, usually workspace/experiment/subject/evokeds.
    params : dict
        Contains additional information about the evoked.
    content : dict of mne.Evoked, optional
        Stores mne.Evoked objects as values, conditions as keys. If not provided,
        is assumed to be saved to file system earlier.
    """

    def __init__(self, name, directory, params, content=None):
        self._name = name.strip('.fif')
        self._path = os.path.join(directory, name + '.fif')
        self._params = params

        # ensure comments are set to match the keys / conditions
        self._content = content
        if self._content:
            for key in self._content.keys():
                self._content[key].comment = key

        # for backwards compatbility,
        # evokeds used to be stored in epochs/average
        if 'bwc_path' in self._params:
            self._bwc_path = os.path.join(self._params.pop('bwc_path'),
                                          self._name + '.fif')

    @property
    def content(self):
        """Returns the actual mne.Evoked objects either
        from cache or from the file system.
        """
        if self._content:
            return self._content

        self._content = {}

        # for backwards compatibility,
        # try first from the new path,
        # and then from the old path
        try:
            evokeds = mne.read_evokeds(self._path)
        except Exception:
            try:
                evokeds = mne.read_evokeds(self._bwc_path)
            except Exception:
                raise IOError('Reading evokeds failed.')

        for evoked in evokeds:
            self._content[evoked.comment] = evoked

        return self._content

    @property
    def name(self):
        """Returns name of the evoked object.
        """
        return self._name

    @property
    def params(self):
        """Returns additional information stored.
        """
        return self._params

    @property
    def ch_names(self):
        """Returns names of the data channels, must read the
        evoked to memory.
        """
        return list(self.content.values())[0].info['ch_names']

    @property
    def times(self):
        """Returns times (array of points in time), must read
        the evoked to memory.
        """
        return list(self.content.values())[0].times

    @property
    def info(self):
        """Returns info structure containing e.g. sensor locations,
        must read the evoked to memory.
        """
        return list(self.content.values())[0].info

    @property
    def data(self):
        """Returns dict of the data (numpy arrays).
        """
        data = {}
        for key in self.content.keys():
            data[key] = self.content[key].data
        return data

    def save_content(self):
        """Saves the mne.Evoked to a fif file in the evoked directory.
        """
        try:
            mne.write_evokeds(self._path, list(self.content.values()))
        except Exception as exc:
            raise Exception("Writing evokeds failed. Please check that the "
                            "entire experiment folder has write permissions.")

    def delete_content(self):
        """Deletes the fif file from the file system.
        """
        os.remove(self._path)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'evokedTopomapDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_evokedTopomapDialog(object):
    def setupUi(self, evokedTopomapDialog):
        evokedTopomapDialog.setObjectName("evokedTopomapDialog")
        evokedTopomapDialog.resize(406, 540)
        self.gridLayout = QtWidgets.QGridLayout(evokedTopomapDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(evokedTopomapDialog)
        self.scrollArea.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 386, 489))
        self.scrollAreaWidgetContents.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBoxSettings = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxSettings.setObjectName("groupBoxSettings")
        self.formLayout = QtWidgets.QFormLayout(self.groupBoxSettings)
        self.formLayout.setObjectName("formLayout")
        self.labelStart = QtWidgets.QLabel(self.groupBoxSettings)
        self.labelStart.setObjectName("labelStart")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.labelStart)
        self.doubleSpinBoxStart = QtWidgets.QDoubleSpinBox(self.groupBoxSettings)
        self.doubleSpinBoxStart.setMinimum(-99.0)
        self.doubleSpinBoxStart.setSingleStep(0.05)
        self.doubleSpinBoxStart.setProperty("value", -0.2)
        self.doubleSpinBoxStart.setObjectName("doubleSpinBoxStart")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxStart)
        self.labelEnd = QtWidgets.QLabel(self.groupBoxSettings)
        self.labelEnd.setObjectName("labelEnd")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.labelEnd)
        self.doubleSpinBoxEnd = QtWidgets.QDoubleSpinBox(self.groupBoxSettings)
        self.doubleSpinBoxEnd.setMinimum(-99.0)
        self.doubleSpinBoxEnd.setSingleStep(0.05)
        self.doubleSpinBoxEnd.setProperty("value", 0.5)
        self.doubleSpinBoxEnd.setObjectName("doubleSpinBoxEnd")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxEnd)
        self.labelStep = QtWidgets.QLabel(self.groupBoxSettings)
        self.labelStep.setObjectName("labelStep")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.labelStep)
        self.doubleSpinBoxStep = QtWidgets.QDoubleSpinBox(self.groupBoxSettings)
        self.doubleSpinBoxStep.setMaximum(1.0)
        self.doubleSpinBoxStep.setSingleStep(0.05)
        self.doubleSpinBoxStep.setProperty("value", 0.1)
        self.doubleSpinBoxStep.setObjectName("doubleSpinBoxStep")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxStep)
        self.checkBoxRadius = QtWidgets.QCheckBox(self.groupBoxSettings)
        self.checkBoxRadius.setObjectName("checkBoxRadius")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.checkBoxRadius)
        self.doubleSpinBoxRadius = QtWidgets.QDoubleSpinBox(self.groupBoxSettings)
        self.doubleSpinBoxRadius.setEnabled(False)
        self.doubleSpinBoxRadius.setDecimals(3)
        self.doubleSpinBoxRadius.setMaximum(10.0)
        self.doubleSpinBoxRadius.setSingleStep(0.01)
        self.doubleSpinBoxRadius.setProperty("value", 0.15)
        self.doubleSpinBoxRadius.setObjectName("doubleSpinBoxRadius")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBoxRadius)
        self.gridLayout_2.addWidget(self.groupBoxSettings, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 1, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.pushButtonCancel = QtWidgets.QPushButton(evokedTopomapDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonApply = QtWidgets.QPushButton(evokedTopomapDialog)
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.horizontalLayout.addWidget(self.pushButtonApply)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(evokedTopomapDialog)
        self.pushButtonCancel.clicked.connect(evokedTopomapDialog.reject)
        self.pushButtonApply.clicked.connect(evokedTopomapDialog.accept)
        self.checkBoxRadius.toggled['bool'].connect(self.doubleSpinBoxRadius.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(evokedTopomapDialog)

    def retranslateUi(self, evokedTopomapDialog):
        _translate = QtCore.QCoreApplication.translate
        evokedTopomapDialog.setWindowTitle(_translate("evokedTopomapDialog", "Meggie - Evoked topomaps"))
        self.groupBoxSettings.setTitle(_translate("evokedTopomapDialog", "Settings:"))
        self.labelStart.setText(_translate("evokedTopomapDialog", "Start:"))
        self.doubleSpinBoxStart.setSuffix(_translate("evokedTopomapDialog", "s"))
        self.labelEnd.setText(_translate("evokedTopomapDialog", "End:"))
        self.doubleSpinBoxEnd.setSuffix(_translate("evokedTopomapDialog", "s"))
        self.labelStep.setText(_translate("evokedTopomapDialog", "Step:"))
        self.doubleSpinBoxStep.setSuffix(_translate("evokedTopomapDialog", "s"))
        self.checkBoxRadius.setText(_translate("evokedTopomapDialog", "Radius:"))
        self.pushButtonCancel.setText(_translate("evokedTopomapDialog", "Cancel"))
        self.pushButtonApply.setText(_translate("evokedTopomapDialog", "Apply"))




""" Defines Spectrum class, stores numpy arrays representing PSDs.
"""

import os
import re
import logging

import numpy as np
import mne

import meggie.utilities.filemanager as filemanager

from meggie.utilities.datatype import Datatype


class Spectrum(Datatype):
    """ Wraps numpy arrays of PSDs.

    MNE-python does not have a dedicated class for storing PSDs. However, for
    our purposes it is good to have similar interface as the evokeds, TFRs,
    etc. Data is stored in csv files.

    Parameters
    ----------
    name : str
        Name of the spectrum, used in the UI lists and in the .exp file.
    directory : str
        Absolute path to the data folder, usually workspace/experiment/subject/spectrums.
    params : dict
        Contains additional information about the spectrum.
    content : dict of np.array, optional
        The spectral content as a numpy array. If not provided,
        is assumed to be saved to file system earlier.
    freqs : np.array, optional
        Frequencies as a array of points in frequency. If not provided,
        is assumed to be saved to file system earlier.
    info : mne.Info, optional
        The info structure from the raw data. Is stored because contains
        channel names and locations. If not provided, is assumed 
        to be saved to file system earlier.

    """
    def __init__(self, name, directory, params,
                 content=None, freqs=None, info=None):
        # name has no group number and no '.fif'
        self._name = name
        self._directory = directory
        self._params = params

        self._content = {}
        if content is not None:
            self._content = content

        self._params['info_set'] = True

        self._freqs = freqs
        self._info = info

    def _load_content(self):
        """Gets content from the file system and 
        stores it to corresponding attributes."""
        data_dict, freqs, ch_names = self._get_content()
        info = self._get_info()
        self._info = info
        self._freqs = freqs
        self._content = data_dict

    def _get_info(self):
        """ Gets info from file system.
        """
        info_path = os.path.join(self._directory,
                                 self._name + '-info.fif')
        info = mne.io.meas_info.read_info(info_path)

        return info

    def _get_content(self):
        """Handles the file loading."""

        # load data
        data_dict = {}
        template = self.name + '_' + r'([a-zA-Z1-9_]+)\.csv'
        for fname in os.listdir(self._directory):
            match = re.match(template, fname)
            if match:
                try:
                    key = str(match.group(1))
                except Exception as exc:
                    raise Exception("Unknown file name format.")

                # if proper condition parameters set,
                # check if the key is in there.
                if 'conditions' in self._params:
                    if key not in [str(elem) for elem in
                                   self._params['conditions']]:
                        continue

                freqs, row_descs, psd = filemanager.load_csv(
                    os.path.join(self._directory, fname))

                ch_names = [desc[0] for desc in row_descs]

                # for backwards compatibility
                # (used to have possibility to have spectrum data
                # saved as log transformed)
                if 'log_transformed' in self._params:
                    if self._params['log_transformed'] is True:
                        if np.mean(psd) < 0:
                            psd = 10 ** (psd / 10.0)

                freqs = np.array(freqs).astype(np.float)
                data_dict[key] = np.array(psd)

        return data_dict, freqs, ch_names

    def save_content(self):
        """Saves spectral data and info structure to the spectrum directory.
        """
        try:
            # save info
            info_path = os.path.join(self._directory,
                                     self._name + '-info.fif')
            mne.io.meas_info.write_info(info_path, self._info)
            self._params['info_set'] = True

            # save data
            for key, psd in self._content.items():

                row_descs = [(ch_name,) for ch_name in self.ch_names]
                column_names = self._freqs.tolist()
                data = psd.tolist()

                path = os.path.join(self._directory,
                                    self._name + '_' + str(key) + '.csv')

                filemanager.save_csv(path, data, column_names, row_descs)
        except Exception as exc:
            raise Exception("Writing spectrums failed. Please check that the "
                            "entire experiment folder has write permissions.")

    def delete_content(self):
        """Removes spectral data and info structure from the
        file system.
        """

        # delete info
        info_path = os.path.join(self._directory,
                                 self._name + '-info.fif')
        if os.path.exists(info_path):
            os.remove(info_path)

        # delete data
        template = self.name + '_' + r'([a-zA-Z1-9_]+)\.csv'
        for fname in os.listdir(self._directory):
            match = re.match(template, fname)
            if match:
                try:
                    key = str(match.group(1))
                except Exception as exc:
                    continue

                # if proper condition parameters set,
                # check if the key is in there.
                if 'conditions' in self._params:
                    if key not in [str(elem) for elem in
                                   self._params['conditions']]:
                        continue

                os.remove(os.path.join(self._directory, fname))

    def set_info(self, subject):
        """Stores info structure to the spectrum object. 

        This is for backwards compatibility. We used to get sensor locations
        from the raw object. This was problematic as the raw could change
        after creation of the spectrum.
        """
        info = subject.get_raw(preload=False).info

        # filter to correct set of channels
        _, _, ch_names = self._get_content()
        picks = [ch_idx for ch_idx, ch_name in enumerate(info['ch_names'])
                 if ch_name in ch_names]
        info = mne.pick_info(info, sel=picks)

        self._info = info
        self.save_content()

    @property
    def data(self):
        """Returns the dict of numpy arrays (conditions as keys, PSDs as values).
        """
        return self.content

    @property
    def content(self):
        """Returns the dict of numpy arrays (conditions as keys, PSDs as values).
        """
        if not self._content:
            self._load_content()
        return self._content

    @property
    def freqs(self):
        """Returns freqs, must read the data to memory first.
        """
        if not self._content:
            self._load_content()
        return self._freqs

    @property
    def ch_names(self):
        """Returns channel names from the info structure, must
        read the data to memory first."""
        return self.info['ch_names']

    @property
    def info(self):
        """Returns the info structure, must read the data to
        memory first."""
        if not self._content:
            self._load_content()
        return self._info

    @property
    def name(self):
        """Returns name of the spectrum"""
        return self._name

    @property
    def params(self):
        """Returns additional information stored, for example
        the conditions that are looked for in the spectrums
        directory, when loading the data.
        """
        return self._params




""" Contains create spectrum action handling.
"""

from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.utilities.dialogs.powerSpectrumDialogMain import PowerSpectrumDialog
from meggie.actions.spectrum_create.controller.spectrum import create_power_spectrum


class CreateSpectrum(Action):
    """ Shows a dialog to get parameters for spectrum creation and
    then allows creating spectrums """

    def run(self):
        default_name = next_available_name(
            self.experiment.active_subject.spectrum.keys(), 'Spectrum')

        dialog = PowerSpectrumDialog(self.experiment, self.window, 
                                     default_name, self.handler)
        dialog.show()


    @subject_action
    def handler(self, subject, params):
        """
        """
        spectrum_name = params['name']
        intervals = params['intervals']
        fmin = params['fmin']
        fmax = params['fmax']
        nfft = params['nfft']
        overlap = params['overlap']
        
        create_power_spectrum(subject, spectrum_name, intervals,
                              fmin, fmax, nfft, overlap,
                              do_meanwhile=self.window.update_ui)




""" Contains a class for the tfr output options dialog.
"""
from PyQt5 import QtWidgets

from meggie.utilities.dialogs.TFROutputOptionsUi import Ui_TFROutputOptions


class TFROutputOptions(QtWidgets.QDialog):
    """ Contains logic for the tfr output options dialog.
    """

    def __init__(self, parent, experiment, tfr_name, handler, ask_condition=False):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_TFROutputOptions()
        self.ui.setupUi(self)
        self.parent = parent
        self.experiment = experiment
        self.handler = handler
        self.ask_condition = ask_condition

        if not ask_condition:
            self.ui.groupBoxCondition.hide()

        active_subject = self.experiment.active_subject

        meggie_tfr = active_subject.tfr[tfr_name]

        tfr = list(meggie_tfr.content.values())[0]
        keys = list(meggie_tfr.content.keys())

        start, end = tfr.times[0], tfr.times[-1]
        minfreq, maxfreq = tfr.freqs[0], tfr.freqs[-1]

        self.ui.doubleSpinBoxBaselineStart.setMinimum(start)
        self.ui.doubleSpinBoxBaselineStart.setMaximum(end)
        self.ui.doubleSpinBoxBaselineEnd.setMinimum(start)
        self.ui.doubleSpinBoxBaselineEnd.setMaximum(end)
        self.ui.doubleSpinBoxTimeStart.setMinimum(start)
        self.ui.doubleSpinBoxTimeStart.setMaximum(end)
        self.ui.doubleSpinBoxTimeEnd.setMinimum(start)
        self.ui.doubleSpinBoxTimeEnd.setMaximum(end)
        self.ui.doubleSpinBoxFrequencyMin.setMinimum(minfreq)
        self.ui.doubleSpinBoxFrequencyMax.setMaximum(maxfreq)
        self.ui.doubleSpinBoxFrequencyMin.setMinimum(minfreq)
        self.ui.doubleSpinBoxFrequencyMax.setMaximum(maxfreq)
 
        self.ui.doubleSpinBoxBaselineStart.setValue(start)
        self.ui.doubleSpinBoxBaselineEnd.setValue(0)
        self.ui.doubleSpinBoxTimeStart.setValue(start)
        self.ui.doubleSpinBoxTimeEnd.setValue(end)
        self.ui.doubleSpinBoxFrequencyMin.setValue(minfreq)
        self.ui.doubleSpinBoxFrequencyMax.setValue(maxfreq)

        for key in keys:
            self.ui.comboBoxCondition.addItem(key)

        if len(keys) == 1:
            self.ui.comboBoxCondition.setEnabled(False)

    def accept(self):
        subject = self.experiment.active_subject

        if self.ask_condition:
            condition = self.ui.comboBoxCondition.currentText()
        else:
            condition = None

        if self.ui.checkBoxBaselineCorrection.isChecked():
            blmode = self.ui.comboBoxBaselineMode.currentText()
        else:
            blmode = None

        blstart = self.ui.doubleSpinBoxBaselineStart.value()
        blend = self.ui.doubleSpinBoxBaselineEnd.value()
        tmin = self.ui.doubleSpinBoxTimeStart.value()
        tmax = self.ui.doubleSpinBoxTimeEnd.value()
        fmin = self.ui.doubleSpinBoxFrequencyMin.value()
        fmax = self.ui.doubleSpinBoxFrequencyMax.value()

        if self.ui.radioButtonAllChannels.isChecked():
            output = 'all_channels'
        else:
            output = 'channel_averages'

        params = {}
        params['output_option'] = output
        params['condition'] = condition
        params['blmode'] = blmode
        params['blstart'] = blstart
        params['blend'] = blend
        params['tmin'] = tmin
        params['tmax'] = tmax
        params['fmin'] = fmin
        params['fmax'] = fmax

        self.handler(params)

        self.close()



""" Contains plot spectrum action handling.
"""

from meggie.utilities.names import next_available_name
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.spectrum_plot.controller.spectrum import plot_spectrum_topo
from meggie.actions.spectrum_plot.controller.spectrum import plot_spectrum_averages

from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class PlotSpectrum(Action):
    """ Plots spectrums.
    """

    def run(self):
        try:
            selected_name = self.data['outputs']['spectrum'][0]
        except IndexError as exc:
            return

        def option_handler(selected_option):
            params = {'name': selected_name, 
                      'output_option': selected_option,
                      'channel_groups': self.experiment.channel_groups}
            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = OutputOptions(self.window, handler=option_handler)
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """
        """
        if params['output_option'] == 'channel_averages':
            plot_spectrum_averages(subject, params['channel_groups'], params['name'])
        else:
            info = subject.get_raw().info
            chs = list(get_channels_by_type(info).keys())
            if 'eeg' in chs:
                plot_spectrum_topo(subject, params['name'], ch_type='eeg')
            if 'grad' in chs or 'mag' in chs:
                plot_spectrum_topo(subject, params['name'], ch_type='meg')




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'eventsFromAnnotationsDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_EventsFromAnnotationsDialog(object):
    def setupUi(self, EventsFromAnnotationsDialog):
        EventsFromAnnotationsDialog.setObjectName("EventsFromAnnotationsDialog")
        EventsFromAnnotationsDialog.resize(406, 540)
        self.gridLayout = QtWidgets.QGridLayout(EventsFromAnnotationsDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(EventsFromAnnotationsDialog)
        self.scrollArea.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollArea.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.scrollArea.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 372, 655))
        self.scrollAreaWidgetContents.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.groupBoxConfiguration = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxConfiguration.setObjectName("groupBoxConfiguration")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.groupBoxConfiguration)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.labelAnnotation = QtWidgets.QLabel(self.groupBoxConfiguration)
        self.labelAnnotation.setObjectName("labelAnnotation")
        self.gridLayout_3.addWidget(self.labelAnnotation, 0, 0, 1, 1)
        self.labelLocation = QtWidgets.QLabel(self.groupBoxConfiguration)
        self.labelLocation.setObjectName("labelLocation")
        self.gridLayout_3.addWidget(self.labelLocation, 1, 0, 1, 1)
        self.pushButtonAdd = QtWidgets.QPushButton(self.groupBoxConfiguration)
        self.pushButtonAdd.setObjectName("pushButtonAdd")
        self.gridLayout_3.addWidget(self.pushButtonAdd, 8, 0, 1, 2)
        self.radioButtonEnd = QtWidgets.QRadioButton(self.groupBoxConfiguration)
        self.radioButtonEnd.setObjectName("radioButtonEnd")
        self.gridLayout_3.addWidget(self.radioButtonEnd, 2, 1, 1, 1)
        self.radioButtonStart = QtWidgets.QRadioButton(self.groupBoxConfiguration)
        self.radioButtonStart.setChecked(True)
        self.radioButtonStart.setObjectName("radioButtonStart")
        self.gridLayout_3.addWidget(self.radioButtonStart, 1, 1, 1, 1)
        self.comboBoxAnnotation = QtWidgets.QComboBox(self.groupBoxConfiguration)
        self.comboBoxAnnotation.setObjectName("comboBoxAnnotation")
        self.gridLayout_3.addWidget(self.comboBoxAnnotation, 0, 1, 1, 1)
        self.labelEventID = QtWidgets.QLabel(self.groupBoxConfiguration)
        self.labelEventID.setObjectName("labelEventID")
        self.gridLayout_3.addWidget(self.labelEventID, 3, 0, 1, 1)
        self.spinBoxEventID = QtWidgets.QSpinBox(self.groupBoxConfiguration)
        self.spinBoxEventID.setMinimum(1)
        self.spinBoxEventID.setMaximum(1000000)
        self.spinBoxEventID.setObjectName("spinBoxEventID")
        self.gridLayout_3.addWidget(self.spinBoxEventID, 3, 1, 1, 1)
        self.listWidgetItems = QtWidgets.QListWidget(self.groupBoxConfiguration)
        self.listWidgetItems.setObjectName("listWidgetItems")
        self.gridLayout_3.addWidget(self.listWidgetItems, 10, 0, 1, 2)
        self.pushButtonClear = QtWidgets.QPushButton(self.groupBoxConfiguration)
        self.pushButtonClear.setObjectName("pushButtonClear")
        self.gridLayout_3.addWidget(self.pushButtonClear, 11, 0, 1, 2)
        self.gridLayout_2.addWidget(self.groupBoxConfiguration, 0, 0, 1, 1)
        self.groupBoxBatching = QtWidgets.QGroupBox(self.scrollAreaWidgetContents)
        self.groupBoxBatching.setObjectName("groupBoxBatching")
        self.gridLayoutBatching = QtWidgets.QGridLayout(self.groupBoxBatching)
        self.gridLayoutBatching.setObjectName("gridLayoutBatching")
        self.batchingWidgetPlaceholder = QtWidgets.QWidget(self.groupBoxBatching)
        self.batchingWidgetPlaceholder.setMinimumSize(QtCore.QSize(300, 300))
        self.batchingWidgetPlaceholder.setObjectName("batchingWidgetPlaceholder")
        self.gridLayoutBatching.addWidget(self.batchingWidgetPlaceholder, 0, 0, 1, 1)
        self.gridLayout_2.addWidget(self.groupBoxBatching, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 2, 0, 1, 1)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.pushButtonCancel = QtWidgets.QPushButton(EventsFromAnnotationsDialog)
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.horizontalLayout.addWidget(self.pushButtonCancel)
        self.pushButtonBatch = QtWidgets.QPushButton(EventsFromAnnotationsDialog)
        self.pushButtonBatch.setObjectName("pushButtonBatch")
        self.horizontalLayout.addWidget(self.pushButtonBatch)
        self.pushButtonApply = QtWidgets.QPushButton(EventsFromAnnotationsDialog)
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.horizontalLayout.addWidget(self.pushButtonApply)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(EventsFromAnnotationsDialog)
        self.pushButtonCancel.clicked.connect(EventsFromAnnotationsDialog.reject)
        self.pushButtonApply.clicked.connect(EventsFromAnnotationsDialog.accept)
        self.pushButtonBatch.clicked.connect(EventsFromAnnotationsDialog.acceptBatch)
        QtCore.QMetaObject.connectSlotsByName(EventsFromAnnotationsDialog)

    def retranslateUi(self, EventsFromAnnotationsDialog):
        _translate = QtCore.QCoreApplication.translate
        EventsFromAnnotationsDialog.setWindowTitle(_translate("EventsFromAnnotationsDialog", "Meggie - Events from annotations"))
        self.groupBoxConfiguration.setTitle(_translate("EventsFromAnnotationsDialog", "Configuration:"))
        self.labelAnnotation.setText(_translate("EventsFromAnnotationsDialog", "Annotation:"))
        self.labelLocation.setText(_translate("EventsFromAnnotationsDialog", "Use:"))
        self.pushButtonAdd.setText(_translate("EventsFromAnnotationsDialog", "Add to list"))
        self.radioButtonEnd.setText(_translate("EventsFromAnnotationsDialog", "End"))
        self.radioButtonStart.setText(_translate("EventsFromAnnotationsDialog", "Start"))
        self.labelEventID.setText(_translate("EventsFromAnnotationsDialog", "Event id:"))
        self.pushButtonClear.setText(_translate("EventsFromAnnotationsDialog", "Clear"))
        self.groupBoxBatching.setTitle(_translate("EventsFromAnnotationsDialog", "Batching:"))
        self.pushButtonCancel.setText(_translate("EventsFromAnnotationsDialog", "Cancel"))
        self.pushButtonBatch.setText(_translate("EventsFromAnnotationsDialog", "Batch"))
        self.pushButtonApply.setText(_translate("EventsFromAnnotationsDialog", "Apply"))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'actionDialogUi.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ActionDialog(object):
    def setupUi(self, ActionDialog):
        ActionDialog.setObjectName("ActionDialog")
        ActionDialog.resize(518, 440)
        self.gridLayoutDialog = QtWidgets.QGridLayout(ActionDialog)
        self.gridLayoutDialog.setObjectName("gridLayoutDialog")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pushButtonClose = QtWidgets.QPushButton(ActionDialog)
        self.pushButtonClose.setObjectName("pushButtonClose")
        self.horizontalLayout.addWidget(self.pushButtonClose)
        self.gridLayoutDialog.addLayout(self.horizontalLayout, 1, 0, 1, 1)
        self.treeWidgetActions = QtWidgets.QTreeWidget(ActionDialog)
        self.treeWidgetActions.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        self.treeWidgetActions.setAnimated(True)
        self.treeWidgetActions.setObjectName("treeWidgetActions")
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidgetActions)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_0 = QtWidgets.QTreeWidgetItem(self.treeWidgetActions)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        item_1 = QtWidgets.QTreeWidgetItem(item_0)
        self.treeWidgetActions.topLevelItem(1).child(2).setText(0, "create_spectrum")
        item_1.setToolTip(0, "")
        item_1.setStatusTip(0, "")
        item_1.setWhatsThis(0, "")
        item_1.setFlags(QtCore.Qt.ItemIsEnabled)
        self.treeWidgetActions.header().setVisible(False)
        self.gridLayoutDialog.addWidget(self.treeWidgetActions, 0, 0, 1, 1)

        self.retranslateUi(ActionDialog)
        self.pushButtonClose.clicked.connect(ActionDialog.accept)
        QtCore.QMetaObject.connectSlotsByName(ActionDialog)

    def retranslateUi(self, ActionDialog):
        _translate = QtCore.QCoreApplication.translate
        ActionDialog.setWindowTitle(_translate("ActionDialog", "Meggie - Actions"))
        self.pushButtonClose.setText(_translate("ActionDialog", "Close"))
        self.treeWidgetActions.headerItem().setText(0, _translate("ActionDialog", "Action"))
        __sortingEnabled = self.treeWidgetActions.isSortingEnabled()
        self.treeWidgetActions.setSortingEnabled(False)
        self.treeWidgetActions.topLevelItem(0).setText(0, _translate("ActionDialog", "subject_001"))
        self.treeWidgetActions.topLevelItem(0).child(0).setText(0, _translate("ActionDialog", "ica"))
        self.treeWidgetActions.topLevelItem(0).child(1).setText(0, _translate("ActionDialog", "filter"))
        self.treeWidgetActions.topLevelItem(1).setText(0, _translate("ActionDialog", "subject_002"))
        self.treeWidgetActions.topLevelItem(1).child(0).setText(0, _translate("ActionDialog", "ica"))
        self.treeWidgetActions.topLevelItem(1).child(1).setText(0, _translate("ActionDialog", "filter"))
        self.treeWidgetActions.setSortingEnabled(__sortingEnabled)



""" Contains classes and functions for experiments.
"""

import os
import json
import importlib
import shutil
import logging
import pkg_resources

from meggie.subject import Subject

from meggie.utilities.filemanager import open_raw
from meggie.utilities.filemanager import save_raw
from meggie.utilities.channels import get_default_channel_groups
from meggie.utilities.validators import validate_name
from meggie.utilities.uid import generate_uid

from meggie.mainwindow.dynamic import find_all_datatype_specs


class Experiment:
    """ A top-level class that contains structure of an experiment,
    all the subjects within, and so forth.

    Parameters
    ----------
    name : str
        Name of the experiment
    author : str
        Name of the author
    path : str
        Path to the experiment folder.
    """
    def __init__(self, name, author, path):
        self._name = name
        self._author = author
        self._path = path
        self._subjects = {}
        self._active_subject = None

        self._channel_groups = {
            'eeg': {},
            'meg': {},
        }
        self._selected_pipeline = 'classic'

    @property
    def path(self):
        """ Returns the path. """
        return self._path

    @path.setter
    def path(self, path):
        """ Sets the path. """
        self._path = path

    @property
    def name(self):
        """ Returns the name. """
        return self._name

    @name.setter
    def name(self, name):
        """ Sets the name. """
        self._name = validate_name(
            name, fieldname='name')

    @property
    def author(self):
        """ Returns the author."""
        return self._author

    @author.setter
    def author(self, author):
        """ Sets the author."""
        self._author = validate_name(author, minlength=0, fieldname='author')

    @property
    def channel_groups(self):
        """ Returns channel groups for experiment. If not set,
        uses defaults."""
        channel_groups = self._channel_groups.copy()

        # if channel groups not found, use defaults..
        if not channel_groups.get('eeg'):
            if self.active_subject:
                raw = self.active_subject.get_raw(preload=False)
                try:
                    channel_groups['eeg'] = get_default_channel_groups(raw.info, 'eeg')
                except Exception as exc:
                    pass

        if not channel_groups.get('meg'):
            if self.active_subject:
                raw = self.active_subject.get_raw(preload=False)
                try:
                    channel_groups['meg'] = get_default_channel_groups(raw.info, 'meg')
                except Exception as exc:
                    pass

        return channel_groups

    @channel_groups.setter
    def channel_groups(self, channel_groups):
        """ Sets the channel groups."""
        self._channel_groups = channel_groups

    @property
    def selected_pipeline(self):
        """ Returns selected pipeline
        """
        return self._selected_pipeline

    @selected_pipeline.setter
    def selected_pipeline(self, selected_pipeline):
        """ Sets selected pipeline """
        self._selected_pipeline = selected_pipeline

    @property
    def active_subject(self):
        """ Returns the active subject.
        """
        return self._active_subject

    @active_subject.setter
    def active_subject(self, subject):
        """ Sets the active subject.
        """
        self._active_subject = subject

    def add_subject(self, subject):
        """ Adds subject to the experiment.

        Parameters
        ----------
        subject : meggie.subject.Subject
            A subject object.
        """
        self.subjects[subject.name] = subject

    @property
    def subjects(self):
        """ Returns contained subjects.
        """
        return self._subjects

    def remove_subject(self, name):
        """Removes a subject.

        Removes the subject folder and its contents under experiment tree, and
        removes the subject from the experiment.

        Parameters
        ----------
        name : str
            Name of the subject to be removed.
        
        """

        if getattr(self, 'active_subject',
                   None) and self.active_subject.name == name:
            self.active_subject = None

        subject = self.subjects.pop(name)

        try:
            shutil.rmtree(subject.path)
        except Exception as exc:
            raise Exception(
                'Could not remove the contents of the subject folder.')

    def activate_subject(self, subject_name):
        """Activates a subject from the existing subjects

        Parameters
        ----------
        subject_name : str
            Name of the subject.

        Returns
        -------
        meggie.subject.Subject
            The activated subject.
        """
        # remove raw files from memory before activating new subject.
        if self.active_subject:
            self.active_subject.release_memory()

        self.active_subject = self.subjects[subject_name]

        # test validity
        self.active_subject.get_raw(preload=False)

        return self.active_subject

    def create_subject(self, subject_name, raw_path):
        """ Creates a subject object and copies the raw file
        inside it (by reading and saving).

        Parameters
        ----------
        subject_name : str
            Name of the new subject.
        raw_path : str
            Path to the data file.
        """
        bname = os.path.basename(raw_path)
        stem, ext = os.path.splitext(bname)
        new_fname = stem + '.fif'

        uid = generate_uid()

        subject = Subject(self, subject_name, new_fname, uid)
        subject.ensure_folders()

        raw = open_raw(raw_path)

        new_path = os.path.join(subject.path, new_fname)
        save_raw(raw, new_path)

        self.add_subject(subject)

    def save_experiment_settings(self):
        """
        Saves the experiment settings into a file in the root of
        the experiment directory structure.
        """
        subjects = []

        for subject in self.subjects.values():
            subject_dict = {
                'subject_name': subject.name,
                'raw_fname': subject.raw_fname,
                'uid': subject.uid,
                'ica_applied': subject.ica_applied,
                'rereferenced': subject.rereferenced,
            }

            datatype_specs = find_all_datatype_specs()

            for source, package, datatype_spec in datatype_specs.values():
                save_key = datatype_spec['save_key']
                datatype = datatype_spec['id']

                for inst in getattr(subject, datatype).values():
                    datatype_dict = {
                        'name': inst.name,
                        'params': inst.params
                    }

                    if save_key not in subject_dict.keys():
                        subject_dict[save_key] = []

                    subject_dict[save_key].append(datatype_dict)

            subjects.append(subject_dict)

        save_dict = {
            'subjects': subjects,
            'name': self._name,
            'author': self._author,
            'channel_groups': self._channel_groups,
            'selected_pipeline': self._selected_pipeline
        }

        try:
            version = pkg_resources.get_distribution("meggie").version
        except BaseException:
            version = ''

        save_dict['version'] = version

        try:
            os.makedirs(self.path)
        except OSError:
            pass

        path = os.path.join(self._path, os.path.basename(self._path) + '.exp')

        # let's backup previous exp file with version number
        try:
            if os.path.exists(path):
                with open(path, 'r') as f:
                    old_data = json.load(f)

                version = str(old_data['version']).replace('.', '-')

                backup_path = os.path.join(self._path, 
                    os.path.basename(self._path) + '_' + version + '.exp.bak')

                shutil.copy(path, backup_path)
        except Exception as exc:
            message = ("Could not backup experiment file to {0}. "
                       "Please check that the experiment folder "
                       "has write permissions everywhere.")
            raise Exception(message.format(backup_path))

        # and then overwrite the current exp file
        try:
            with open(path, 'w') as f:
                json.dump(save_dict, f, sort_keys=True, indent=4)
        except Exception as exc:
            message = ("Could not save experiment file {0}. "
                       "Please check that the experiment folder "
                       "has write permissions everywhere.")
            raise Exception(message.format(path))

def initialize_new_experiment(name, author, prefs, set_previous_experiment=True):
    """Initializes new experiment object with given data.

    Parameters
    ----------
    name : str
        Name of the experiment.
    author : str
        Name of the author.
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.

    Returns
    -------
    meggie.experiment.Experiment
        The new experiment.

    """
    path = os.path.join(prefs.workspace, name)
    if os.path.exists(path):
        raise Exception('Experiment with same name already exists.')

    experiment = Experiment(name, author, path)
    experiment.save_experiment_settings()

    if set_previous_experiment:
        prefs.previous_experiment_name = experiment.path
        prefs.write_preferences_to_disk()
    return experiment


def open_existing_experiment(prefs, path=None):
    """ Reads and opens existing experiment.

    Parameters
    ----------
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.
    path : str, optional
        Path to the experiment folder.

    Returns
    -------
    meggie.experiment.Experiment
        The opened experiment.
    """

    experiment_updated = False

    if path:
        exp_file = os.path.join(path, os.path.basename(path) + '.exp')
    else:
        exp_file = os.path.join(
            prefs.previous_experiment_name,
            os.path.basename(prefs.previous_experiment_name) + '.exp'
        )

    if not os.path.isfile(exp_file):
        raise ValueError(
            'Trying to open experiment without settings file (.exp).')

    try:
        with open(exp_file, 'r') as f:
            data = json.load(f)
    except ValueError as exc:
        logging.getLogger('ui_logger').exception('')
        raise ValueError('Experiment from ' + exp_file + ' could not be ' +
                         'opened. There might be a problem with ' +
                         'cohesion of the experiment file.')

    if not path:
        path = os.path.dirname(exp_file)

    experiment = Experiment(data['name'], data['author'], path)

    if 'channel_groups' in data.keys() and data['channel_groups'] != 'MNE':
        experiment.channel_groups = data['channel_groups']
    else:
        experiment.channel_groups = {
            'eeg': {},
            'meg': {},
        }

    if 'selected_pipeline' in data.keys():
        experiment.selected_pipeline = data['selected_pipeline']

    if len(data['subjects']) == 0:
        return experiment

    for subject_data in data['subjects']:

        subject_name = subject_data['subject_name']

        raw_fname = subject_data.get('raw_fname')
        # for backwards compatibility
        if not raw_fname:
            raw_fname = subject_data.get('working_file_name')
        if not raw_fname:
            raise Exception('raw_fname not set in the exp file')

        uid = subject_data.get('uid')
        if not uid:
            uid = generate_uid()
            experiment_updated = True

        subject = Subject(experiment,
                          subject_name,
                          raw_fname,
                          uid,
                          ica_applied=subject_data.get('ica_applied', False),
                          rereferenced=subject_data.get('rereferenced', False)
                          )

        datatype_specs = find_all_datatype_specs()

        for source, package, datatype_spec in datatype_specs.values():
            save_key = datatype_spec['save_key']
            entry = datatype_spec['entry']
            datatype = datatype_spec['id']

            for inst_data in subject_data.get(save_key, []):
                module_name, class_name = entry.split('.')
                module = importlib.import_module(
                    '.'.join([source, 'datatypes', package, module_name]))
                inst_class = getattr(module, class_name, None)

                name = inst_data.get('name')
                # backward compatibility
                if not name and 'collection_name' in inst_data:
                    name = inst_data.get('collection_name')
                if not name:
                    raise Exception('No name attribute found')

                directory = getattr(subject, datatype + '_directory')
                params = inst_data.get('params', {})

                # for backwards compatibility
                if datatype == 'evoked':
                    if not params.get(
                            'conditions') and 'event_names' in inst_data:
                        params['conditions'] = inst_data['event_names']
                    params['bwc_path'] = os.path.join(
                        subject.path, 'epochs/average')
                if datatype == 'spectrum':
                    if not params.get(
                            'log_transformed') and 'log_transformed' in inst_data:
                        params['log_transformed'] = inst_data['log_transformed']
                if datatype == 'tfr':
                    if not params.get(
                            'decim') and 'decim' in inst_data:
                        params['decim'] = inst_data['decim']
                    if not params.get(
                            'n_cycles') and 'n_cycles' in inst_data:
                        params['n_cycles'] = inst_data['n_cycles']
                    if not params.get(
                            'evoked_subtracted') and 'evoked_subtracted' in inst_data:
                        params['evoked_subtracted'] = inst_data['evoked_subtracted']

                if inst_class:
                    inst = inst_class(
                        name,
                        directory,
                        params.copy(),
                    )
                    subject.add(inst, datatype)

                    # for backwards compatibility,
                    # ensure spectrum objects contain info
                    if datatype == 'spectrum' and not params.get('info_set'):
                        inst.set_info(subject)

        experiment.add_subject(subject)

        # ensure that the folder structure exists
        # (to not crash on updates)
        subject.ensure_folders()

    if experiment_updated:
        experiment.save_experiment_settings()

    return experiment



""" Contains implementation for evoked topomaps
"""
import logging

import matplotlib.pyplot as plt
import numpy as np
import mne

from matplotlib.gridspec import GridSpec

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.evoked_plot_topomap.dialogs.evokedTopomapDialogMain import EvokedTopomapDialog


class PlotEvokedTopomap(Action):
    """ Plots a sequence of topomaps.
    """

    def run(self):

        try:
            selected_name = self.data['outputs']['evoked'][0]
        except IndexError as exc:
            return

        subject = self.experiment.active_subject
        evoked = subject.evoked.get(selected_name)

        def handle_close(tmin, tmax, step, radius):
            params = {}
            params['tmin'] = tmin
            params['tmax'] = tmax
            params['step'] = step
            params['radius'] = radius
            params['name'] = selected_name

            self.handler(subject, params)

        dialog = EvokedTopomapDialog(self.window, evoked, handle_close)
        dialog.show()


    @subject_action
    def handler(self, subject, params):
        name = params['name']
        tmin = params['tmin']
        tmax = params['tmax']
        step = params['step']
        radius = params['radius']
        evoked = subject.evoked.get(name)

        for key, evok in sorted(evoked.content.items()):
            channels = get_channels_by_type(evok.info)
            for ch_type in channels.keys():
                title_elems = [name, key, ch_type]
                times = np.arange(tmin, tmax, step)

                # Use custom figure so that mne does not remove the mpl toolbar
                fig = plt.figure()
                axes = []

                # one subplot for each topomap
                for idx in range(len(times)):
                    spec = GridSpec(5, 2*(len(times)+1)).new_subplotspec((1, 2*idx),
                                                                     rowspan=3, colspan=2)
                    axes.append(fig.add_subplot(spec))

                # and one for colorbar
                spec = GridSpec(5, 2*(len(times)+1)).new_subplotspec((1, 2*len(times)),
                                                                 rowspan=3, colspan=1)
                axes.append(fig.add_subplot(spec))

                # until there is a good solution to topomap skirts, fix a value
                sphere = None
                if ch_type in ['mag', 'grad']:
                    sphere = radius

                fig = mne.viz.plot_evoked_topomap(
                    evok, times=times, ch_type=ch_type,
                    title=' '.join(title_elems), axes=axes, sphere=sphere)
                fig.canvas.set_window_title('_'.join(title_elems))




""" Contains controlling logic for the evoked plot
"""

import mne
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D

from meggie.utilities.plotting import color_cycle
from meggie.utilities.plotting import create_channel_average_plot
from meggie.utilities.channels import average_to_channel_groups

from meggie.utilities.units import get_unit


def _create_averages(mne_evoked, channel_groups):
    mne_evoked = mne_evoked.copy().drop_channels(mne_evoked.info['bads'])

    data_labels, averaged_data = average_to_channel_groups(
        mne_evoked.data, mne_evoked.info, 
        mne_evoked.info['ch_names'], channel_groups)

    return data_labels, averaged_data


def plot_evoked_averages(evoked, channel_groups):
    """ Plots channel averages.
    """
    conditions = evoked.content.keys()
    colors = color_cycle(len(conditions))
    times = evoked.times

    averages = {}
    for key, mne_evoked in sorted(evoked.content.items()):
        data_labels, averaged_data = _create_averages(mne_evoked, channel_groups)

        for label_idx, label in enumerate(data_labels):
            if not label in averages:
                averages[label] = []
            averages[label].append((key, averaged_data[label_idx]))

    ch_types = sorted(set([label[0] for label in averages.keys()]))
    for ch_type in ch_types:

        ch_groups = sorted([label[1] for label in averages.keys() 
                            if label[0] == ch_type])

        def plot_fun(ax_idx, ax):
            ch_group = ch_groups[ax_idx]
            ax.set_title(ch_group)

            ax.set_xlabel('Time (s)')
            ax.set_ylabel('Amplitude ({})'.format(
                get_unit(ch_type)))

            for color_idx, (key, curve) in enumerate(averages[(ch_type, ch_group)]):
                ax.plot(times, curve, color=colors[color_idx])

            ax.axhline(0, color='black')
            ax.axvline(0, color='black')

        title = ' '.join([evoked.name, ch_type])
        legend = list(zip(conditions, colors))
        create_channel_average_plot(len(ch_groups), plot_fun, title, 
                                    legend)

    plt.show()


def plot_evoked_topo(evoked, ch_type):
    """ Plots evoked time courses arranged as a topography """
    evokeds = []
    labels = []
    for key, evok in sorted(evoked.content.items()):

        info = evok.info
        if ch_type == 'eeg':
            dropped_names = [ch_name for ch_idx, ch_name in enumerate(info['ch_names'])
                             if ch_idx not in mne.pick_types(info, eeg=True, meg=False)]
        else:
            dropped_names = [ch_name for ch_idx, ch_name in enumerate(info['ch_names'])
                             if ch_idx not in mne.pick_types(info, eeg=False, meg=True)]

        evok = evok.copy().drop_channels(dropped_names)

        evokeds.append(evok)
        labels.append(key)

    colors = color_cycle(len(evoked.content.keys()))

    # setup legend for subplots
    lines = [Line2D([0], [0], color=colors[idx], label=labels[idx])
             for idx in range(len(labels))]
    def onclick(event):
        try:
            # not nice:
            ax = plt.gca()

            channel = plt.getp(ax, 'title')
            ax.set_title('')

            ax.legend(handles=lines, loc='upper right')

            title = ' '.join([evoked.name, channel])
            ax.figure.suptitle(title)
            ax.figure.canvas.set_window_title(title.replace(' ', '_'))
            plt.show()
        except Exception as exc:
            pass

    fig = mne.viz.plot_evoked_topo(evokeds, color=colors)
    fig.canvas.mpl_connect('button_press_event', onclick)
    title = "{0}_{1}".format(evoked.name, ch_type)
    fig.canvas.set_window_title(title)




""" Contains implementation for events from annotations
"""
import logging

import matplotlib.pyplot as plt
import numpy as np
import mne

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.threading import threaded

from meggie.actions.raw_events_from_annotations.dialogs.eventsFromAnnotationsDialogMain import EventsFromAnnotationsDialog

from meggie.utilities.events import events_from_annotations

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class EventsFromAnnotations(Action):
    """ Shows a dialog for parameter selection and then
    applies a conversion from annotations to events.
    """

    @subject_action
    def handler(self, subject, params):
        """
        """
        @threaded
        def conv_fun():
            """
            """
            events_from_annotations(subject, params['items'])
            subject.save()

        conv_fun(do_meanwhile=self.window.update_ui)

    def run(self):
        evs_from_annots_dialog = EventsFromAnnotationsDialog(
            self.window, self.experiment, self.handler)
        evs_from_annots_dialog.show()



""" Contains a class for logic of experiment creation dialog.
"""

import os
import pkg_resources
import json

from PyQt5 import QtWidgets

from meggie.mainwindow.dialogs.createExperimentDialogUi import Ui_CreateExperimentDialog

from meggie.experiment import initialize_new_experiment

from meggie.mainwindow.dynamic import find_all_package_specs

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox


class CreateExperimentDialog(QtWidgets.QDialog):
    """ Contains logic for experiment creation dialog.
    """

    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self, parent)

        self.ui = Ui_CreateExperimentDialog()
        self.ui.setupUi(self)

        self.parent = parent
        prefs = parent.prefs

        self.active_plugins = prefs.active_plugins

        # read all pipeline ids and names to a list
        pipelines = []
        package_specs = find_all_package_specs()

        for source, package_spec in package_specs.items():

            if source == 'meggie' or source in self.active_plugins:
                if 'pipelines' in package_spec:
                    for pipeline in package_spec['pipelines']:
                        try:
                            id_ = pipeline['id']
                        except Exception as exc:
                            raise Exception('Every pipeline should have id.')

                        name = pipeline.get('name', '')
                        pipelines.append((id_, name))

        # Add classic
        pipelines.append(('classic', 'Include everything'))

        self.pipelines = pipelines

        # Create buttons for pipelines
        self.pipeline_buttons = []
        for idx, (pipeline_id, pipeline_name) in enumerate(self.pipelines):
            radio_button = QtWidgets.QRadioButton(self.ui.groupBoxPipeline)
            radio_button.setText(pipeline_name)

            self.ui.gridLayoutPipeline.addWidget(
                radio_button, idx + 1, 0, 1, 1)

            self.pipeline_buttons.append(radio_button)

            if idx == 0:
                radio_button.setChecked(True)

        if len(self.pipeline_buttons) == 1:
            self.pipeline_buttons[0].setEnabled(False)
            self.pipeline_buttons[0].setChecked(True)

    def accept(self):
        if self.ui.lineEditExperimentName.text() == '':
            message = 'Give experiment a name.'
            messagebox(self.parent, message)
            return

        selected_pipeline = ""
        for button_idx, radio_button in enumerate(self.pipeline_buttons):
            if radio_button.isChecked():
                selected_pipeline = self.pipelines[button_idx][0]
                break

        try:
            experiment = initialize_new_experiment(
                self.ui.lineEditExperimentName.text(),
                self.ui.lineEditAuthor.text(),
                self.parent.prefs
            )

            experiment.selected_pipeline = selected_pipeline
            experiment.save_experiment_settings()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.experiment = experiment
        self.parent.reconstruct_tabs()
        self.parent.initialize_ui()
        self.close()



""" Contains implementation for spectrum info
"""
from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float


class Info(InfoAction):
    """ Fills up spectrum info box """

    def run(self):
        try:
            selected_name = self.data['outputs']['spectrum'][0]

            spectrum = self.experiment.active_subject.spectrum[selected_name]
            params = spectrum.params

            message = ""

            message += "Name: "+ spectrum.name + "\n\n"

            if 'fmin' in params and 'fmax' in params:
                message += 'Frequencies: {0}Hz - {1}Hz\n'.format(format_float(params['fmin']),
                                                                 format_float(params['fmax']))

            if 'nfft' in params:
                message += 'Window length (samples): {0}\n'.format(params['nfft'])


            if 'overlap' in params:
                message += 'Overlap (samples): {0}\n'.format(params['overlap'])

            if 'intervals' in params:
                message += '\nIntervals: \n'
                for key, ivals in params['intervals'].items():
                    message += 'Condition ' + str(key) + ': '
                    message += ', '.join(['({0}s - {1}s)'.format(format_float(ival[0]), format_float(ival[1]))
                                          for ival in ivals])
                    message += '\n'

            if 'groups' in params:
                for key, names in params['groups'].items():
                    message += '\nGroup ' + str(key) + ': \n'
                    for name in names:
                        message += name + '\n'

        except Exception as exc:
            message = ""

        return message
